from __future__ import print_function
from __future__ import absolute_import

import string
import os
import weakref
import re
import glob
import numpy as np
import pandas as pds
from pysat import data_dir as data_dir

class Files(object):
    """Maintains collection of files for instrument object.
    
    Uses the list_files functions for each specific instrument
    to create an ordered collection of files in time. Used by
    instrument object to load the correct files. Files also
    contains helper methods for determining the presence of
    new files and creating an ordered list of files.
      
    Attributes
    ----------
    base_path : string
        path to .pysat directory in user home
    start_date : datetime
        date of first file, used as default start bound for instrument
        object
    stop_date : datetime
        date of last file, used as default stop bound for instrument
        object
    data_path : string
        path to the directory containing instrument files,
        top_dir/platform/name/tag/
        
    Note
    ----
    User should generally use the interface provided by a pysat.Instrument
    instance. Exceptions are the classmethod from_os, provided to assist
    in generating the appropriate output for an instrument routine.
    
    Examples
    --------
    ::
    
        # convenient file access
        inst = pysat.Instrument(platform=platform, name=name, tag=tag)
        # first file
        inst.files[0]
        
        # files from start up to stop (exclusive on stop)
        start = pysat.datetime(2009,1,1)
        stop = pysat.datetime(2009,1,3)
        print(vefi.files[start:stop])
        
        # files for date
        print(vefi.files[start])
        
        # files by slicing
        print(vefi.files[0:4])
        
        # get a list of new files
        # new files are those that weren't present the last time
        # a given instrument's file list was stored
        new_files = vefi.files.get_new()
        
        # search pysat appropriate directory for instrument files and
        # update Files instance, knowledge not written to disk.
        vefi.files.refresh()
        
        # search pysat appropriate directory for files and store new list
        vefi.files.refresh(store=True)
        # running get_new will now return an empty list until
        # additional files are introduced
        
        
    """
        
    def __init__(self, sat):

        self._sat = weakref.proxy(sat)
        self.base_path = os.path.join(os.getenv('HOME'), '.pysat')
        self.start_date = None
        self.stop_date = None
        self.files = pds.Series(None)
        self.data_path = os.path.join(data_dir, self._sat.platform, 
                                      self._sat.name, self._sat.tag)
        
        if self._sat.platform != '':
            info = self._load()
            if info is not False:
                self._attach_files(info)
            else:
                print("pysat is searching for the requested instrument's files.")
                # couldn't find stored info, load file list and then store
                info = self._sat._list_rtn(tag=self._sat.tag, data_path=self.data_path)
                if not info.empty:
                    info = self._remove_data_dir_path(info)	
                    self._attach_files(info)
                    self._store()

    def _attach_files(self, files_info):
        """Attaches info returned by instrument list_files routine to Instrument object."""

        if (len(files_info.index.unique()) != len(files_info)):
            print('Duplicate datetimes ', files_info.index.get_duplicates())
            raise ValueError('List of files must have unique datetimes.')

        self.files = files_info.sort_index()
        date = files_info.index[0]
        self.start_date = pds.datetime(date.year, date.month, date.day)
        date = files_info.index[-1]
        self.stop_date = pds.datetime(date.year, date.month, date.day)

    def _store(self, dir=None):
        """Store currently loaded filelist for instrument onto filesystem"""
        name = ''.join((self._sat.platform,'_',self._sat.name,'_',self._sat.tag,
                        '_stored_file_info.txt'))
        if dir is None:
            name = os.path.join(self.base_path, name)
        else:
            name = os.path.join(dir, name)
        try:
            self.files.to_csv(name, date_format='%Y-%m-%d %H:%M:%S.%f')
            return True
        except IOError:
            return False	

    def _load(self):
        """Load stored filelist and return as Pandas Series"""
        fname = ''.join((self._sat.platform,'_',self._sat.name,'_',
                        self._sat.tag, '_stored_file_info.txt'))
        fname = os.path.join(self.base_path, fname)
        try:
            data = pds.Series.from_csv(fname, index_col=0)
        except IOError:
            return False
        else:
            return data

    def refresh(self, store=False):
        """Refresh loaded instrument filelist by searching filesystem.
        
        Searches pysat provided path, pysat_data_dir/platform/name/tag/, 
        where pysat_data_dir is set by pysat.utils.set_data_dir(path=path).
        
        Parameters
        ----------
        store : boolean
            set True to store loaded file names into .pysat directory
            
        """
        info = self._sat._list_rtn(tag=self._sat.tag, data_path=self.data_path)
        info = self._remove_data_dir_path(info)
        self._attach_files(info)
        if store:
            self._store()

    def get_new(self):
        """List all new files since last time list was stored.
        
        pysat stores filenames in the user_home/.pysat directory. Returns
        a list of all new fileanmes since the last store. Filenames
        are stored if update_files is True at instrument object level and
        if files.refresh(store=True) is called.
        
        Returns
        -------
            pandas Series of filenames
            False if no filenames
        """
        stored_info = self._load()
        if stored_info is not False:
            new_info = self._sat._list_rtn(tag = self._sat.tag, data_path=self.data_path)
            new_info = self._remove_data_dir_path(new_info)
            new_files = new_info[~new_info.isin(stored_info) ]
            return new_files
        else:
            print('No previously stored files that we may compare to.')
            return False

    def get_index(self, fname):
        """Return index for a given filename. 
        
        Parameters
        ----------
        fname : string
            filename
        
        Note
        ----
        If fname not found in the file information already attached 
        to the instrument.files instance, then a files.refresh() call 
        is made.
        
        """
        idx, = np.where(fname == self.files)
        if len(idx) == 0:
            # filename not in index, try reloading files from disk
            self.refresh()
            print (fname, self.files)
            idx, = np.where(fname == self.files)
            if len(idx) == 0:
                raise IOError('Could not find supplied file on disk')
        return idx

    # convert this to a normal get so files[in:in2] gives the same as requested here
    # support slicing via date and index
    # filename is inclusive slicing, date and index are normal non-inclusive end point

    def __getitem__(self, key):
        if isinstance(key, slice):
            try:
                out = self.files.ix[key]
            except IndexError:
                raise IndexError('Date requested outside file bounds.')
            if isinstance(key.start, pds.datetime):
                # enforce exclusive slicing on datetime
                if len(out) > 1:
                    if out.index[-1] >= key.stop:
                        return out[:-1]
                    else:
                        return out
                elif len(out) == 1:
                    if out.index[0] >= key.stop:
                        return pds.Series([])
                    else:
                        return out
                else:
                    return out
            else:
                # not a datetime
                return out
        else:
            return self.files.ix[key]
            #raise ValueError('Not implemented yet.')         
        #if isinstance(key, tuple):
        #    if len(key) == 2:
        #        start = key[0]
        #        end = key[1]
        #    else:
        #        raise ValueError('Must input 2 and only 2 items/iterables')
                
    def get_file_array(self, start, end):
        """Return a list of filenames between and including start and end.
        
        Parameters
        ----------
            start: array_like or single string 
                filenames for start of returned filelist
            stop: array_like or single string
                filenames inclusive end of list
                
        Returns
        -------
            list of filenames between and including start and end over all
            intervals. 
            
        """
        if hasattr(start, '__iter__') & hasattr(end, '__iter__'):
            files = []
            for (sta,stp) in zip(start, end):
                id1 = self.get_index(sta)
                id2 = self.get_index(stp)
                files.extend(self.files.iloc[id1[0] : id2[0]+1])
        elif hasattr(start, '__iter__') | hasattr(end, '__iter__'):
            raise ValueError('Either both or none of the inputs need to be iterable')
        else:
            id1 = self.get_index(start)
            id2 = self.get_index(end)
            files = self.files[id1[0]:id2[0]+1].to_list()
        return files

    def _remove_data_dir_path(self, inp=None):
        # import string
        """Remove the data directory path from filenames"""
        # need to add a check in here to make sure data_dir path is actually in
        # the filename
        if inp is not None:
            split_str = os.path.join(self.data_path, '')
            return inp.apply(lambda x: x.split(split_str)[-1])
            # match = os.path.join(self.data_path,'')
            # num = len(match)
            # return inp.apply(lambda x: x[num:])
        
    @classmethod    
    def from_os(cls, data_path=None, format_str=None, 
                two_digit_year_break=None):
        """
        Produces a list of files and and formats it for Files class.
        
        Parameters
        ----------
            data_path : string
                Top level directory to search files for. This directory
                is provided by pysat to the instrument_module.list_files
                functions as data_path.
            format_string : string with python format codes
                Provides the naming pattern of the instrument files and the 
                locations of date information so an ordered list may be 
                produced.
            two_digit_year_break : int
                If filenames only store two digits for the year, then
                '1900' will be added for years >= two_digit_year_break,
                and '2000' will be added for years < two_digit_year_break.
                
        Note
        ----
        Does not produce a Files instance, but the proper output
        from instrument_module.list_files method.
        
        """
        import collections
        
        from pysat.utils import create_datetime_index
        
        if format_str is None:
            raise ValueError("Must supply a filename template (format_str).")
        if data_path is None:
            raise ValueError("Must supply instrument directory path (dir_path)")
        
        # parse format string to figure out the search string to use
        # to identify files in the filesystem
        search_str = ''
        form = string.Formatter()
        keys = []
        snips = []
        length = []
        stored = collections.OrderedDict()
        stored['year'] = []; stored['month'] = []; stored['day'] = [];
        stored['hour'] = []; stored['min'] = []; stored['sec'] = [];
        for snip in form.parse(format_str):
            search_str += snip[0]
            snips.append(snip[0])
            if snip[1] is not None:
                keys.append(snip[1])
                search_str += '*'
                # try and determine formatting width
                temp = re.findall(r'\d+', snip[2])
                if temp:
                    # there are items, try and grab width
                    for i in temp:
                        if i != 0:
                            length.append(int(i))
                            break
                else:
                    raise ValueError("Couldn't determine formatting width")

        abs_search_str = os.path.join(data_path, search_str)
        files = glob.glob(abs_search_str)   
        
        # we have a list of files, now we need to extract the date information        
        # code below works, but only if the size of file string 
        # remains unchanged
        
        # determine the loactaions the date information in a filename is stored
        # use these indices to slice out date from loaded filenames
        # test_str = format_str.format(**periods)
        if len(files) > 0:  
            idx = 0
            begin_key = []
            end_key = []
            for i,snip in enumerate(snips):
                idx += len(snip)
                if i < (len(length)):
                    begin_key.append(idx)
                    idx += length[i]
                    end_key.append(idx)
            max_len = idx
            # setting up negative indexing to pick out filenames
            key_str_idx = [np.array(begin_key, dtype=int) - max_len, 
                            np.array(end_key, dtype=int) - max_len]
            # need to parse out dates for datetime index
            for i,temp in enumerate(files):
                for j,key in enumerate(keys):
                    val = temp[key_str_idx[0][j]:key_str_idx[1][j]]
                    stored[key].append(val)
            # convert to numpy arrays
            for key in stored.keys():
                stored[key] = np.array(stored[key]).astype(int)
                if len(stored[key]) == 0:
                    stored[key]=None
            # deal with the possibility of two digit years
            # years above or equal to break are considered to be 1900+
            # years below break are considered to be 2000+
            if two_digit_year_break is not None:
                idx, = np.where(stored['year'] >= two_digit_year_break)
                stored['year'][idx] = stored['year'][idx] + 1900
                idx, = np.where(stored['year'] < two_digit_year_break)
                stored['year'][idx] = stored['year'][idx] + 2000 
            # need to sort the information for things to work
            rec_arr = [stored[key] for key in keys]
            rec_arr.append(files)
            # sort all arrays
            val_keys = keys+['files']
            rec_arr = np.rec.fromarrays(rec_arr, names=val_keys)
            rec_arr.sort(order=val_keys, axis=0)
            # pull out sorted info
            for key in keys:
                stored[key] = rec_arr[key]
            files = rec_arr['files']
            # add hour and minute information to 'sec'
            if stored['sec'] is None:
                stored['sec'] = np.zeros(len(files))                
            if stored['hour'] is not None:
                stored['sec'] += 3600*stored['hour']
            if stored['min'] is not None:
                stored['sec'] += 60*stored['min']
            
            index = create_datetime_index(year=stored['year'], month=stored['month'], 
                                    day=stored['day'], uts=stored['sec'])

            return pds.Series(files, index=index)
        else:
            print ("Unable to find any files. If you have the necessary files "+
                   "please check pysat settings and file locations.")
            return pds.Series(None) 




# -*- coding: utf-8 -*-
"""SuperDARN data support for grdex files(Alpha Level!)

Parameters
----------
platform : string
    'superdarn'
name : string
    'grdex'
tag : string
    'north' or 'south' for Northern/Southern hemisphere data

Note
----
Requires davitpy and pydarn to load SuperDARN files.
Uses environment variables set by davitpy to download files
from Virginia Tech SuperDARN data servers. Pydarn routines
are used to load SuperDARN data.

Warnings
--------
Cleaning only removes entries that have 0 vectors, grdex files
are constituted from what it is thought to be good data.
"""
from __future__ import print_function
from __future__ import absolute_import
import sys
import os

import pandas as pds
import numpy as np

import pysat
import pydarn

def list_files(tag=None, data_path=None):
    """Return a Pandas Series of every file for chosen satellite data"""

    if tag is not None:
        if tag == 'north':
            return pysat.Files.from_os(data_path=data_path, 
                format_str='{year:4d}{month:02d}{day:02d}.north.grdex')
        else:
            raise ValueError('Unrecognized tag name for SuperDARN, north or south.')                  
    else:
        raise ValueError ('A tag name must be passed to SuperDARN.')           
                
           
def load_orig(fnames, tag=None):
    if len(fnames) <= 0 :
        return pysat.DataFrame(None), pysat.Meta(None)
    elif len(fnames)==1:
        b = pydarn.sdio.sdDataOpen(pysat.datetime(1980,1,1), 
                                    src='local', 
                                    eTime=pysat.datetime(2050,1,1),
                                    fileName=fnames[0])
                                    
        data_list = pydarn.sdio.sdDataReadAll(b)
        sys.stdout.flush()
        in_dict = []
        for info in data_list:
            arr = np.arange(len(info.stid))
            drift_frame = pds.DataFrame(info.vector.__dict__, 
                                    #index=[info.vector.mlon, info.vector.mlat])
                                    index=info.vector.index)
            drift_frame.index.name = 'index'
            drift_frame.sort(inplace=True)
            #drift_frame.index.names=['mlon', 'mlat']
            for i in arr:
                nvec = info.nvec[i]
                in_frame = drift_frame.iloc[0:nvec]
                drift_frame = drift_frame.iloc[nvec:]
                in_dict.append({'stid':info.stid[i],
                            'channel':info.channel[i],
                            'noisemean':info.noisemean[i],
                            'noisesd':info.noisesd[i],
                            'gsct':info.gsct[i],
                            'nvec':info.nvec[i],
                            'pmax':info.pmax[i],
                            'vector':in_frame,
                            'start_time':info.sTime,
                            'end_time':info.eTime,
                            'vemax':info.vemax[i],
                            'vemin':info.vemin[i],
                            'pmin':info.pmin[i],
                            'programid':info.programid[i],
                            'wmax':info.wmax[i],
                            'wmin':info.wmin[i],
                            'freq':info.freq[i]})
        output = pds.DataFrame(in_dict)
        output.index = output.start_time
        output.drop('start_time', axis=1, inplace=True)
        return output, pysat.Meta()
    else:
        raise ValueError('Only one filename currently supported.')

def load(fnames, tag=None):
    if len(fnames) <= 0 :
        return pysat.DataFrame(None), pysat.Meta(None)
    elif len(fnames)==1:
        
        #b = pydarn.sdio.sdDataOpen(pysat.datetime(1980,1,1), 
        #                            src='local', 
        #                            eTime=pysat.datetime(2050,1,1),
        #                            fileName=fnames[0])
        
        myPtr = pydarn.sdio.sdDataPtr(sTime=pysat.datetime(1980,1,1),
                          eTime=pysat.datetime(2250,1,1),
                          hemi=tag)  
        myPtr.fType, myPtr.dType = 'grdex', 'dmap'                 
        myPtr.ptr = open(fnames[0],'r')
                                                                                            
        #data_list = pydarn.sdio.sdDataReadAll(myPtr)
        in_list = []
        in_dict = {'stid':[],
            'channel':[],
            'noisemean':[],
            'noisesd':[],
            'gsct':[],
            'nvec':[],
            'pmax':[],
            'start_time':[],
            'end_time':[],
            'vemax':[],
            'vemin':[],
            'pmin':[],
            'programid':[],
            'wmax':[],
            'wmin':[],
            'freq':[]}
        
        #for info in data_list:
        while True:
            info = pydarn.dmapio.readDmapRec(myPtr.ptr)   
            info = pydarn.sdio.sdDataTypes.gridData(dataDict=info)
            if info.channel is None:
                break 
                   
            drift_frame = pds.DataFrame.from_records(info.vector.__dict__, 
                                                     nrows=len(info.pmax),
                                                     index=info.vector.index)
            drift_frame.index.name = 'index'
            sum_vec = 0
            for nvec in info.nvec:
                #in_frame = drift_frame.iloc[0:nvec]
                #drift_frame = drift_frame.iloc[nvec:]
                in_list.append(drift_frame.iloc[sum_vec:sum_vec+nvec])
                sum_vec += nvec

            in_dict['stid'].extend(info.stid)
            in_dict['channel'].extend(info.channel)
            in_dict['noisemean'].extend(info.noisemean)
            in_dict['noisesd'].extend(info.noisesd)
            in_dict['gsct'].extend(info.gsct)
            in_dict['nvec'].extend(info.nvec)
            in_dict['pmax'].extend(info.pmax)
            in_dict['start_time'].extend([info.sTime]*len(info.pmax))
            in_dict['end_time'].extend([info.eTime]*len(info.pmax))
            in_dict['vemax'].extend(info.vemax)
            in_dict['vemin'].extend(info.vemin)
            in_dict['pmin'].extend(info.pmin)
            in_dict['programid'].extend(info.programid)
            in_dict['wmax'].extend(info.wmax)
            in_dict['wmin'].extend(info.wmin)
            in_dict['freq'].extend(info.freq)

        output = pds.DataFrame(in_dict)
        output['vector'] = in_list
        output.index = output.start_time
        output.drop('start_time', axis=1, inplace=True)
        myPtr.ptr.close()
        return output, pysat.Meta()
    else:
        raise ValueError('Only one filename currently supported.')

                                        
#def default(ivm):
#
#    return
            
def clean(self):
    # remove data when there are no vectors
    idx, = np.where(self['nvec'] > 0)
    self.data = self.data.iloc[idx]

    return  
    
def download(date_array, tag, data_path, user=None, password=None):
    """
    download IVM data consistent with pysat

    """

    import sys
    import os
    import pysftp
    
    if user is None:
        user = os.environ['DBREADUSER']
    if password is None:
        password = os.environ['DBREADPASS']
        
    with pysftp.Connection(
            os.environ['VTDB'], 
            username=user,
            password=password) as sftp:
            
        for date in date_array:
            myDir = '/data/'+date.strftime("%Y")+'/grdex/'+tag+'/'
            fname = date.strftime("%Y%m%d")+'.north.grdex'
            local_fname = fname+'.bz2'
            saved_fname = os.path.join(data_path,local_fname) 
            full_fname = os.path.join(data_path,fname) 
            try:
                print('Downloading file for '+date.strftime('%D'))
                sys.stdout.flush()
                sftp.get(myDir+fname, saved_fname)
                os.system('bunzip2 -c '+saved_fname+' > '+ full_fname)
                os.system('rm ' + saved_fname)
            except IOError:
                print('File not available for '+date.strftime('%D'))


         
#
#
#def test3():
#    b = pydarn.sdio.sdDataOpen(pysat.datetime(1980,1,1), 
#                                src='local', 
#                                eTime=pysat.datetime(2050,1,1),
#                                fileName=fnames[0])
#                                
#    data_list = pydarn.sdio.sdDataReadAll(b)
#    print 'building dataframe'
#    sys.stdout.flush()
#    in_frame=[]
#    names=['stid','channel','noisemean','noisesd','gsct','nvec','pmax','vector',
#            'start_time','end_time','vemax','vemin','pmin','programid',
#            'wmax','wmin','freq']
#    
#    in_dict={}    
#    for name in names:
#        in_dict[name]=[]
#
#    for info in data_list:
#
#        drift_frame = pds.DataFrame(info.vector.__dict__, 
#                                index=[info.vector.mlon, info.vector.mlat])
#        drift_frame.index.names=['mlon', 'mlat']
#        in_dict['stid'].extend(info.stid)
#        in_dict['channel'].extend(info.channel)
#        in_dict['noisemean'].extend(info.noisemean)
#        in_dict['noisesd'].extend(info.noisesd)
#        in_dict['gsct'].extend(info.gsct)
#        in_dict['nvec'].extend(info.nvec)
#        in_dict['pmax'].extend(info.pmax)
#        in_dict['start_time'].extend([info.sTime]*len(info.stid))
#        in_dict['end_time'].extend([info.eTime]*len(info.stid))
#        in_dict['vemax'].extend(info.vemax)
#        in_dict['vemin'].extend(info.vemin)
#        in_dict['pmin'].extend(info.pmin)
#        in_dict['programid'].extend(info.programid)
#        in_dict['wmax'].extend(info.wmax)
#        in_dict['wmin'].extend(info.wmin)
#        in_dict['freq'].extend(info.freq)
#        for nvec in info.nvec:
#            in_frame.append(drift_frame.iloc[0:nvec])
#            drift_frame = drift_frame.iloc[nvec:]
#    in_dict['vector'] = in_frame
#    output = pds.DataFrame(in_dict)
#    output.index = output.start_time
#    output.drop('start_time', axis=1, inplace=True) 


from __future__ import print_function
from __future__ import absolute_import

from matplotlib import cm
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np

def scatterplot(inst, labelx, labely, data_label, datalim, xlim=None, ylim=None):
    """Return scatterplot of data_label(s) as functions of labelx,y over a season.

    Parameters
    ----------
    labelx : string 
        data product for x-axis
    labely : string 
        data product for y-axis
    data_label : string, array-like of strings
        data product(s) to be scatter plotted
    datalim : numyp array 
        plot limits for data_label

    Returns
    -------
    Returns a list of scatter plots of data_label as a function
    of labelx and labely over the season delineated by start and
    stop datetime objects.
    
    """

    # interactive plotting off
    plt.ioff()
    
    # create figures for plotting
    figs = []
    axs = []
    
    # check if doing multiple data quantities
    if type(data_label) is str:
        # max of one data type
        figs.append( plt.figure() )
        ax1 = figs[0].add_subplot(211, projection='3d')
        ax2 = figs[0].add_subplot(212)
        axs.append((ax1, ax2))
        plt.suptitle(data_label)
        if xlim is not None:
            ax1.set_xlim(xlim)
        ax2.set_xlim(xlim)
        if ylim is not None:
            ax1.set_ylim(ylim)
        ax2.set_ylim(ylim)
        data_label = [data_label]
        plt.hold(True)
    else:
        # multiple data to be plotted
        for i in np.arange(len(data_label)):
            figs.append( plt.figure() )
            ax1 = figs[i].add_subplot(211, projection='3d')
            ax2 = figs[i].add_subplot(212)
            axs.append((ax1, ax2))
            #axs.append( figs[i].add_subplot(111, projection='3d') )
            plt.suptitle(data_label[i])              
            if xlim is not None:
               ax1.set_xlim(xlim)
            ax2.set_xlim(xlim)
            if ylim is not None:
               ax1.set_ylim(ylim)
            ax2.set_ylim(ylim)
            plt.hold(True)
                
    # norm method so that data may be scaled to colors appropriately
    norm = colors.Normalize(vmin=datalim[0], vmax=datalim[1])    
    p = [i for i in np.arange(len(figs))]
    for i,inst in enumerate(inst):
        for j, (fig, ax) in enumerate(zip(figs, axs)):
            if len(inst.data) > 0:
                if (len(inst.data[labelx]) > 0) & (len(inst.data[labely]) > 0) & (len(inst.data[data_label[j]]) > 0):
                    p[j]=ax[0].scatter(inst.data[labelx], inst.data[labely], inst.data[data_label[j]],
                                       zdir='z', c=inst.data[data_label[j]], cmap=cm.jet, norm=norm,
                                       linewidth='0', edgecolors=None)
                    ax[1].scatter(inst.data[labelx], inst.data[labely], c=inst.data[data_label[j]],
                                  cmap=cm.jet, norm=norm, linewidth=0.00000000001, alpha=0.5, edgecolor=None)
                
    for j, (fig, ax) in enumerate(zip(figs, axs)):
        try:
            cbar = plt.colorbar(p[j],ax=ax[0], label='Amplitude (m/s)')
        except:
            print('Tried colorbar but failed, thus no colorbar.')
        ax[0].elev=30.
    # interactive plotting back on
    plt.ion()
    return figs



from __future__ import print_function
from __future__ import absolute_import

import pandas as pds


class Custom(object):
    """
    Applies a queue of functions when instrument.load called.
    
    Nano-kernel functionality enables instrument objects that are
    'set and forget'. The functions are always run whenever
    the instrument load routine is called so instrument objects may 
    be passed safely to other routines and the data will always
    be processed appropriately.
    
    Examples
    --------
    ::
    
        def custom_func(inst, opt_param1=False, opt_param2=False):
            return None
        instrument.custom.add(custom_func, 'modify', opt_param1=True)
        
        def custom_func2(inst, opt_param1=False, opt_param2=False):
            return data_to_be_added
        instrument.custom.add(custom_func2, 'add', opt_param2=True)
        instrument.load(date=date)
        print(instrument['data_to_be_added'])
    
    See Also
    --------
    Custom.add
    
    Note
    ----
    User should interact with Custom through pysat.Instrument instance's 
    attribute, instrument.custom
    
    """

    def __init__(self):
        # create a list object to store functions
        self._functions = []
        # type of function stored (add/modify/pass on data)
        self._kind = []
        # arguments to functions
        self._args = []
        # keyword arguments to functions
        self._kwargs = []

    def add(self, function, kind='add', at_pos='end',*args, **kwargs):
        """Add a function to custom processing queue. 
        
        Custom functions are applied automatically to associated
        pysat instrument whenever instrument.load command called.

        Parameters
        ----------
            function : string or function object
                name of function or function object to be added to queue
            
            kind : {'add', 'modify', 'pass}
                add 
                    Adds data returned from function to instrument object. 
                    A copy of pysat instrument object supplied to routine. 
                modify  
                    pysat instrument object supplied to routine. Any and all
                    changes to object are retained.
                pass  
                    A copy of pysat object is passed to function. No 
                    data is accepted from return.
                       
            at_pos : string or int 
                insert at position. (default, insert at end).
            args : extra arguments
                extra arguments are passed to the custom function (once)
            kwargs : extra keyword arguments
                extra keyword args are passed to the custom function (once)
                
        Note
        ----
        Allowed `add` function returns:
            
        - {'data' : pandas Series/DataFrame/array_like,
          'units' : string/array_like of strings, 
          'long_name' : string/array_like of strings,
          'name' : string/array_like of strings (iff data array_like)} 
           
        - pandas DataFrame, names of columns are used
        
        - pandas Series, .name required 
        
        - (string/list of strings, numpy array/list of arrays) 
                                                          
        """

        if isinstance(function, str):
            # convert string to function object
            function=eval(function)

        if (at_pos == 'end') | (at_pos == len(self._functions)):
            # store function object
            self._functions.append(function)
            self._args.append(args)
            self._kwargs.append(kwargs)
            self._kind.append(kind.lower())
        elif at_pos < len(self._functions):
            # user picked a specific location to insert
            self._functions.insert(at_pos, function)
            self._args.insert(at_pos, args)
            self._kwargs.insert(at_pos, kwargs)
            self._kind.insert(at_pos, kind)
        else:
            raise TypeError('Must enter an index between 0 and %i' % len(self._functions))

    def _apply_all(self, sat):
        """
        Apply all of the custom functions to the satellite data object.
        """
        if len(self._functions) > 0:
            for func, arg, kwarg, kind in zip(self._functions, self._args, 
                                                self._kwargs, self._kind):
                if len(sat.data) > 0:     
                    if kind == 'add':
                        # apply custom functions that add data to the instrument object
                        tempd = sat.copy()
                        newData = func(tempd, *arg, **kwarg)
                        del tempd
                        
                        # not fancy below, but it works
                        if isinstance(newData,dict):
                            if isinstance(newData['data'], pds.DataFrame):
                                sat[newData['data'].columns] = newData
                            elif isinstance(newData['data'], pds.Series):
                                if newData['data'].name is not None:
                                    sat[newData['data'].name] = newData
                                elif 'name' in newData.keys():
                                    name = newData.pop('name')
                                    sat[name] = newData
                                else:
                                    raise ValueError('Must assign a name to Series'+
                                            ' or return a "name" in dictionary.')
                            elif hasattr(newData['data'], '__iter__'):
                                if 'name' in newData.keys():
                                    name = newData.pop('name')
                                    sat[name] = newData
                                else:
                                    raise ValueError('Must include "name" in returned dictionary.')
                                    
                        elif isinstance(newData, pds.DataFrame):
                            sat[newData.columns] = newData
                        elif isinstance(newData, pds.Series):
                            sat[newData.name] = newData                            
                        elif hasattr(newData, '__iter__'):
                            # falling back to older behavior
                            # unpack tuple/list that was returned
                            newName = newData[0]
                            newData = newData[1]
                            if len(newData)>0:
                                # doesn't really check ensure data, there could
                                # be multiple empty arrays returned, [[],[]]
                                if isinstance(newName, str):
                                    # one item to add
                                    sat[newName] = newData
                                else:    		
                                    # multiple items
                                    for name, data in zip(newName, newData):
                                        if len(data)>0:        
                                            # fixes up the incomplete check from before
                                            sat[name] = data
                        else:
                            raise ValueError("kernel doesn't know what to do with returned data.")
                    # modifying loaded data
                    if kind == 'modify':
                        t = func(sat,*arg,**kwarg)
                        if t is not None:
                            raise ValueError('Modify functions should not return any information via return. Information may only be propagated back by modifying supplied pysat object.')
                    # pass function (function runs, no data allowed back)
                    if kind == 'pass':
                        tempd = sat.copy()
                        t = func(tempd,*arg,**kwarg)
                        del tempd
                        if t is not None:
                            raise ValueError('Pass functions should not return any information via return.')

    def clear(self):
        """Clear custom function list."""
        self._functions=[]
        self._args=[]
        self._kwargs=[]
        self._kind=[]

######################################################
##### END CUSTOM CLASS ##############################
####################################################



from __future__ import print_function
from __future__ import absolute_import

"""Occurrence probability routines, daily or by orbit.

Routines calculate the occurrence of an event greater than a supplied gate
occuring at least once per day, or once per orbit. The probability is
calculated as the (number of times with at least one hit in bin)/(number
of times in the bin).The data used to determine the occurrence must be 1D. 
If a property of a 2D or higher dataset is needed attach a custom function 
that performs the check and returns a 1D Series.

Note
----
The included routines use the bounds attached to the supplied instrument 
object as the season of interest.

"""

import numpy as np

def daily2D(inst, bin1, label1, bin2, label2, data_label, gate, returnBins=False):
    """2D Daily Occurrence Probability of data_label > gate over a season.
    
    If data_label is greater than gate at least once per day, 
    then a 100% occurrence probability results.Season delineated by the bounds
    attached to Instrument object. 
    Prob = (# of times with at least one hit)/(# of times in bin)

    Parameters
    ----------
    inst: pysat.Instrument()
        Instrument to use for calculating occurrence probability
    binx: list
        [min, max, number of bins]
    labelx: string 
        name for data product for binx
    data_label: list of strings 
        identifies data product(s) to calculate occurrence probability
        e.g. inst[data_label]
    gate: list of values 
        values that data_label must achieve to be counted as an occurrence
    returnBins: Boolean
        if True, return arrays with values of bin edges, useful for pcolor

    Returns
    -------
    occur_prob : dictionary
        A dict of dicts indexed by data_label. Each entry is dict with entries
        'prob' for the probability and 'count' for the number of days with any
        data; 'bin_x' and 'bin_y' are also returned if requested. Note that arrays
        are organized for direct plotting, y values along rows, x along columns.
    
    Note
    ----
    Season delineated by the bounds attached to Instrument object.
    
    """
    
    return _occurrence2D(inst, bin1, label1, bin2, label2, data_label, gate,
                        by_orbit=False, returnBins=returnBins)


def by_orbit2D(inst, bin1, label1, bin2, label2, data_label, gate, returnBins=False):
    """2D Occurrence Probability of data_label orbit-by-orbit over a season.
    
    If data_label is greater than gate atleast once per orbit, then a 
    100% occurrence probability results. Season delineated by the bounds
    attached to Instrument object.
    Prob = (# of times with at least one hit)/(# of times in bin)
    
    Parameters
    ----------
    inst: pysat.Instrument()
        Instrument to use for calculating occurrence probability
    binx: list
        [min value, max value, number of bins]
    labelx: string 
        identifies data product for binx
    data_label: list of strings 
        identifies data product(s) to calculate occurrence probability
    gate: list of values 
        values that data_label must achieve to be counted as an occurrence
    returnBins: Boolean
        if True, return arrays with values of bin edges, useful for pcolor

    Returns
    -------
    occur_prob : dictionary
        A dict of dicts indexed by data_label. Each entry is dict with entries
        'prob' for the probability and 'count' for the number of orbits with any
        data; 'bin_x' and 'bin_y' are also returned if requested. Note that arrays
        are organized for direct plotting, y values along rows, x along columns.

    Note
    ----
    Season delineated by the bounds attached to Instrument object.
    
    """
    
    return _occurrence2D(inst, bin1, label1, bin2, label2, data_label, gate,
                        by_orbit=True, returnBins=returnBins)

                                                                                                
def _occurrence2D(inst, bin1, label1, bin2, label2, data_label, gate, 
                by_orbit=False, returnBins=False):
    if not hasattr(data_label, '__iter__'):
        raise ValueError('Data label must be list-like group of variable names.')
    if not hasattr(gate, '__iter__'):
        raise ValueError('Gate levels must be list-like group of variable names.')
    if len(gate) != len(data_label):
        raise ValueError('Must have a gate value for each data_label')

    #create bins
    binx = np.linspace(bin1[0], bin1[1], bin1[2]+1)
    biny = np.linspace(bin2[0], bin2[1], bin2[2]+1)

    numx = len(binx)-1
    numy = len(biny)-1
    numz = len(data_label)
    arrx = np.arange(numx)
    arry = np.arange(numy)
    arrz = np.arange(numz)

    # create arrays to store all values
    total = np.zeros((numz, numy, numx))
    hits = np.zeros((numz, numy, numx))
    if by_orbit:
        inst.load(date=inst.bounds[0][0])
        iterator = inst.orbits
    else:
        iterator = inst

    for i,inst in enumerate(iterator):
        if len(inst.data) != 0:
            xind = np.digitize(inst.data[label1], binx)-1
            for xi in arrx:
                xindex, = np.where(xind==xi)
                if len(xindex)>0:
                    yData = inst.data.iloc[xindex]
                    yind = np.digitize(yData[label2], biny)-1
                    for yj in arry:
                        yindex, = np.where(yind==yj)
                        if len(yindex) > 0:
                            # iterate over the different data_labels
                            for zk in arrz:
                                zdata = yData.ix[yindex,data_label[zk]]
                                if np.any(np.isfinite(zdata)):    
                                    total[zk,yj,xi] += 1.
                                    if np.any(zdata > gate[zk]):
                                        hits[zk,yj,xi] += 1.
                               
    # all of the loading and storing data is done
    # get probability
    prob = hits/total
    # make nicer dictionary output
    output = {}
    for i,label in enumerate(data_label):
        output[label] = {'prob': prob[i,:,:], 'count':total[i,:,:]}
        if returnBins:
            output[label]['bin_x'] = binx
            output[label]['bin_y'] = biny
    # clean up
    del iterator
    return output
        
        
def daily3D(inst, bin1, label1, bin2, label2, bin3, label3, 
            data_label, gate, returnBins=False):
    """3D Daily Occurrence Probability of data_label > gate over a season.
    
    If data_label is greater than gate atleast once per day, 
    then a 100% occurrence probability results. Season delineated by 
    the bounds attached to Instrument object.
    Prob = (# of times with at least one hit)/(# of times in bin)

    Parameters
    ----------
    inst: pysat.Instrument()
        Instrument to use for calculating occurrence probability
    binx: list
        [min, max, number of bins]
    labelx: string 
        name for data product for binx
    data_label: list of strings 
        identifies data product(s) to calculate occurrence probability
    gate: list of values 
        values that data_label must achieve to be counted as an occurrence
    returnBins: Boolean
        if True, return arrays with values of bin edges, useful for pcolor

    Returns
    -------
    occur_prob : dictionary
        A dict of dicts indexed by data_label. Each entry is dict with entries
        'prob' for the probability and 'count' for the number of days with any
        data; 'bin_x', 'bin_y', and 'bin_z' are also returned if requested. Note
        that arrays are organized for direct plotting, z,y,x.

    Note
    ----
    Season delineated by the bounds attached to Instrument object.    
            
    """
        
    return _occurrence3D(inst, bin1, label1, bin2, label2, bin3, label3, 
                        data_label, gate, returnBins=returnBins, by_orbit=False)
           

def by_orbit3D(inst, bin1, label1, bin2, label2, bin3, label3, 
                data_label, gate, returnBins=False):
    """3D Occurrence Probability of data_label orbit-by-orbit over a season.
    
    If data_label is greater than gate atleast once per orbit, then a 
    100% occurrence probability results. Season delineated by the bounds
    attached to Instrument object.
    Prob = (# of times with at least one hit)/(# of times in bin)

    Parameters
    ----------
    inst: pysat.Instrument()
        Instrument to use for calculating occurrence probability
    binx: list
        [min value, max value, number of bins]
    labelx: string 
        identifies data product for binx
    data_label: list of strings 
        identifies data product(s) to calculate occurrence probability
    gate: list of values 
        values that data_label must achieve to be counted as an occurrence
    returnBins: Boolean
        if True, return arrays with values of bin edges, useful for pcolor

    Returns
    -------
    occur_prob : dictionary
        A dict of dicts indexed by data_label. Each entry is dict with entries
        'prob' for the probability and 'count' for the number of orbits with any
        data; 'bin_x', 'bin_y', and 'bin_z' are also returned if requested. Note
        that arrays are organized for direct plotting, z,y,x.

    Note
    ----
    Season delineated by the bounds attached to Instrument object.    
            
    """
        
    return _occurrence3D(inst, bin1, label1, bin2, label2, bin3, label3, 
                        data_label, gate, returnBins=returnBins, by_orbit=True)

                                     
                                                                                                               
def _occurrence3D(inst, start, stop, bin1, label1, bin2, label2, bin3, label3, 
                    data_label, gate, returnBins=False, by_orbit=False):

    if not hasattr(data_label, '__iter__'):
        raise ValueError('Data label must be list-like group of variable names.')
    if not hasattr(gate, '__iter__'):
        raise ValueError('Gate levels must be list-like group of variable names.')
    if len(gate) != len(data_label):
        raise ValueError('Must have a gate value for each data_label')

    #create bins
    binx = np.linspace(bin1[0], bin1[1], bin1[2]+1)
    biny = np.linspace(bin2[0], bin2[1], bin2[2]+1)
    binz = np.linspace(bin3[0], bin3[1], bin3[2]+1)

    numx = len(binx)-1
    numy = len(biny)-1
    numz = len(binz)-1
    numd = len(data_label)

    #create array to store all values before taking median
    yarr = np.arange(numy)
    xarr = np.arange(numx)
    zarr = np.arange(numz)
    darr = np.arange(numd)

    total = np.zeros((numd, numz, numy, numx))
    hits = np.zeros((numd, numz, numy, numx))

    if by_orbit:
        iterator = inst.orbits
    else:
        iterator = inst
    # do loop to iterate over given season
    for i,sat in enumerate(iterator):

        if len(sat.data) != 0:
            xind = np.digitize(sat.data[label1], binx)-1
            for xi in xarr:
                xindex, = np.where(xind==xi)
                if len(xindex)>0:
                    yData = sat.data.ix[xindex]
                    yind = np.digitize(yData[label2], biny)-1
                    for yj in yarr:
                        yindex, = np.where(yind==yj)
                        if len(yindex) > 0:
                            zData = yData.ix[yindex]
                            zind = np.digitize(zData[label3], binz)-1
                            for zk in zarr:
                                zindex, = np.where(zind==zk)
                                if len(zindex) > 0:
                                    for di in darr:
                                        ddata = zData.ix[zindex,data_label[di]]
                                        idx, = np.where(np.isfinite(ddata))
                                        if len(idx) > 0:
                                            total[di,zk,yj,xi] += 1
                                            idx, = np.where( ddata > gate[di]  )
                                            if len(idx) > 0:
                                                hits[di,zk,yj,xi] += 1
                                
                               
    #all of the loading and storing data is done
    #prob = np.zeros((numz, numy, numx))
    prob = hits/total

    # make nicer dictionary output
    output = {}
    for i,label in enumerate(data_label):
        output[label] = {'prob': prob[i,:,:,:], 'count':total[i,:,:,:]}
        if returnBins:
            output[label]['bin_x'] = binx
            output[label]['bin_y'] = biny
            output[label]['bin_z'] = biny
    # clean up
    del iterator
    return output    


# -*- coding: utf-8 -*-
"""Supports OMNI Combined, Definitive, IMF and Plasma Data, and Energetic Proton Fluxes,
Time-Shifted to the Nose of the Earth's Bow Shock, plus Solar and Magnetic Indices. Downloads data from the
NASA Coordinated Data Analysis Web (CDAWeb). Supports both 5 and 1 minute files.

Parameters
----------
platform : string
    'omni'
name : string
    'hro'
tag : string
    Select time between samples, one of {'1min', '5min'}

Note
----
Files are stored by the first day of each month. When downloading use
omni.download(start, stop, freq='MS') to only download days that could possibly have data.
'MS' gives a monthly start frequency.

Warnings
--------
- Currently no cleaning routine. Though the CDAWEB description indicates that these level-2 products
  are expected to be ok.
- Module not written by OMNI team.
        
"""

from __future__ import print_function
from __future__ import absolute_import
import os
import sys

import pandas as pds
import numpy as np

from spacepy import pycdf
import pysat


def list_files(tag=None, data_path=None):
    """Return a Pandas Series of every file for chosen satellite data"""
    if data_path is not None:
        if (tag == '1min') | (tag == '5min'):
            files = pysat.Files.from_os(data_path=data_path,
                    format_str=''.join(['omni_hro_',tag,'{year:4d}{month:02d}{day:02d}_v01.cdf']))
            # files are by month, just repeat filename in a given month for each day of the month
            # load routine will select out appropriate data
            if not files.empty:
                files.ix[files.index[-1]+pds.DateOffset(months=1)-pds.DateOffset(days=1)] = files.iloc[-1]
                files = files.asfreq('D', 'pad')
            return files
        else:
            raise ValueError('Unknown tag')
    else:
        raise ValueError ('A directory must be passed to the loading routine for VEFI')
            

def load(fnames, tag=None):
    if len(fnames) <= 0 :
        return pysat.DataFrame(None), None
    else:
        try:
            cdf = pycdf.CDF(fnames[0])
        except pycdf.CDFError:
            return pysat.DataFrame(), pysat.Meta()
            
        data = {}
        meta = pysat.Meta()
        for key in cdf.iterkeys():
            key_low = key.lower()
            data[key_low] = cdf[key][...]
            try:             
                meta[key_low] = {'units':cdf[key].attrs['UNITS'],
                            'long_name':cdf[key].attrs['LABLAXIS'], 
                            'description':cdf[key].attrs['CATDESC'],
                            'fill_value':cdf[key].attrs['FILLVAL']}
            except KeyError:
                if 'UNITS' in cdf[key].attrs:
                    meta[key_low] = {'units':cdf[key].attrs['UNITS']}
                if 'LABLAXIS' in cdf[key].attrs:
                    meta[key_low] = {'long_name':cdf[key].attrs['LABLAXIS']}
                if 'CATDESC' in cdf[key].attrs:
                    meta[key_low] = {'description':cdf[key].attrs['CATDESC']}
                if 'FILLVAL' in cdf[key].attrs:
                    meta[key_low] = {'fill_value':cdf[key].attrs['FILLVAL']}
                pass
        epoch = data.pop('epoch')
    data = pysat.DataFrame(data, index=pds.to_datetime(epoch, unit='s'))
    return data, meta

def clean(omni):
    for key in omni.data.columns:
        idx, = np.where(omni[key] == omni.meta[key].fill_value)
        omni.data.ix[idx, key] = np.nan

def default(omni):
    """ OMNI data stored monthly. Select out desired day."""

    start = omni.date
    stop = omni.date + pds.DateOffset(days=1) - pds.DateOffset(microseconds=1)
    omni.data = omni.data.ix[start:stop]
    return omni

def download(date_array, tag, data_path=None, user=None, password=None):
    """
    download OMNI data, layout consistent with pysat
    """
    import os
    import ftplib

    ftp = ftplib.FTP('cdaweb.gsfc.nasa.gov')   # connect to host, default port
    ftp.login()               # user anonymous, passwd anonymous@
    
    if (tag == '1min') | (tag == '5min'):
        ftp.cwd('/pub/data/omni/omni_cdaweb/hro_'+tag)
    
        for date in date_array:
            fname = '{year1:4d}/omni_hro_'+tag+'_{year2:4d}{month:02d}{day:02d}_v01.cdf'
            fname = fname.format(year1=date.year, year2=date.year, month=date.month, day=date.day)
            local_fname = ''.join(['omni_hro_',tag,'_{year:4d}{month:02d}{day:02d}_v01.cdf']).format(
                    year=date.year, month=date.month, day=date.day)
            saved_fname = os.path.join(data_path,local_fname) 
            try:
                print('Downloading file for '+date.strftime('%D'))
                sys.stdout.flush()
                ftp.retrbinary('RETR '+fname, open(saved_fname,'w').write)
            except ftplib.error_perm as exception:
                if exception[0][0:3] != '550':
                    raise
                else:
                    os.remove(saved_fname)
                    print('File not available for '+ date.strftime('%D'))



"""
tests the pysat meta object and code
"""
import pysat
import pandas as pds
from nose.tools import assert_raises, raises
import nose.tools
import pysat.instruments.pysat_testing
import numpy as np

class TestBasics:
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.meta = pysat.Meta()
        self.testInst = pysat.Instrument('pysat', 'testing', '', 'clean')

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst
    
    def test_basic_meta_assignment(self):
        self.meta['new'] = {'units':'hey', 'long_name':'boo'}
        assert (self.meta['new'].units == 'hey') & (self.meta['new'].long_name == 'boo')

    def test_replace_meta_units(self):
        self.meta['new'] = {'units':'hey', 'long_name':'boo'}
        self.meta['new'] = {'units':'yep'}
        assert (self.meta['new'].units == 'yep') & (self.meta['new'].long_name == 'boo')

    def test_replace_meta_long_name(self):
        self.meta['new'] = {'units':'hey', 'long_name':'boo'}
        self.meta['new'] = {'long_name':'yep'}
        assert (self.meta['new'].units == 'hey') & (self.meta['new'].long_name == 'yep')
    
    def test_add_additional_metadata_types(self):
        self.meta['new'] = {'units':'hey', 'long_name':'boo', 'description':'boohoo'}

        assert ((self.meta['new'].units == 'hey') & 
                (self.meta['new'].long_name == 'boo') &
                (self.meta['new'].description == 'boohoo'))

    def test_add_meta_then_add_additional_metadata_types(self):
        self.meta['new'] = {'units':'hey', 'long_name':'crew'}
        self.meta['new'] = {'units':'hey', 'long_name':'boo', 'description':'boohoo'}

        assert ((self.meta['new'].units == 'hey') & 
                (self.meta['new'].long_name == 'boo') &
                (self.meta['new'].description == 'boohoo'))
            
    def test_add_meta_then_add_different_additional_metadata_types(self):
        self.meta['new1'] = {'units':'hey1', 'long_name':'crew'}
        self.meta['new2'] = {'units':'hey', 'long_name':'boo', 'description':'boohoo'}
        assert ((self.meta['new2'].units == 'hey') & 
                (self.meta['new2'].long_name == 'boo') &
                (self.meta['new2'].description == 'boohoo') &
                (self.meta['new1'].units == 'hey1') &
                (self.meta['new1'].long_name == 'crew') &
                (np.isnan(self.meta['new1'].description)))

    def test_add_meta_then_partially_add_additional_metadata_types(self):
        self.meta['new'] = {'units':'hey', 'long_name':'crew'}
        self.meta['new'] = {'long_name':'boo', 'description':'boohoo'}

        assert ((self.meta['new'].units == 'hey') & 
                (self.meta['new'].long_name == 'boo') &
                (self.meta['new'].description == 'boohoo'))
                                                                                                                                                                                                                        
    #def test_replace_meta_units_list(self):
    #    self.meta['new'] = {'units':'hey', 'long_name':'boo'}
    #    self.meta['new2'] = {'units':'hey2', 'long_name':'boo2'}
    #    self.meta['new2','new'] = {'units':['yeppers','yep']}
    #    print self.meta['new']
    #    print self.meta['new2']
    #    assert ((self.meta['new'].units == 'yep') & (self.meta['new'].long_name == 'boo') &
    #        (self.meta['new2'].units == 'yeppers') & (self.meta['new2'].long_name == 'boo2'))






# -*- coding: utf-8 -*-
#Test some of the basic _core functions
import sys
import pysat
import pandas as pds
from nose.tools import assert_raises, raises
import nose.tools
import pysat.instruments.pysat_testing
import numpy as np
if sys.version_info[0] >= 3:
    if sys.version_info[1] < 4:
        import imp
        reload = imp.reload
    else:
        import importlib
        reload = importlib.reload


class TestBasics:
    def setup(self):
        reload(pysat.instruments.pysat_testing)
        '''Runs before every method to create a clean testing setup.'''
        self.testInst = pysat.Instrument('pysat', 'testing', '10', 'clean')

    def teardown(self):
        '''Runs after every method to clean up previous testing.'''
        del self.testInst
    
    def test_basic_instrument_load(self):
        '''Test if the correct day is being loaded (checking object date).'''
        self.testInst.load(2009,1)
        assert self.testInst.date == pds.datetime(2009,1,1)	

    def test_next_load_default(self):
        '''Test if first day is loaded by default when first invoking .next.'''
        self.testInst.next()
        assert self.testInst.date == pds.datetime(2008,1,1)

    def test_prev_load_default(self):
        '''Test if last day is loaded by default when first invoking .prev.'''
        self.testInst.prev()
        assert self.testInst.date == pds.datetime(2010,12,31)
    
    def test_filename_load(self):
        '''Test if file is loadable by filename, relative to top_data_dir/platform/name/tag'''
        self.testInst.load(fname='12/31/10.nofile')
        assert self.testInst.data.index[0] == pds.datetime(2010,12,31)
        
    def test_instrument_init(self):
        """Test if init function supplied by instrument can modify object"""
        assert self.testInst.new_thing==True

    @raises(StopIteration)
    def test_left_bounds_with_prev(self):
        '''Test if passing bounds raises StopIteration.'''
        self.testInst.next()
        self.testInst.prev()
        self.testInst.prev()        
        
    @raises(StopIteration)
    def test_right_bounds_with_next(self):
        '''Test if passing bounds raises StopIteration.'''
        self.testInst.prev()
        self.testInst.next()
        self.testInst.next()        

    def test_basic_instrument_load_data(self):
        '''Test if the correct day is being loaded (checking data).'''
        self.testInst.load(2009,1)
        assert self.testInst.data.index[0] == pds.datetime(2009,1,1,0,0,0)

    def test_basic_instrument_load_leap_year(self):
        '''Test if the correct day is being loaded (Leap-Year).'''
        self.testInst.load(2008,366)
        assert self.testInst.date == pds.datetime(2008,12,31)	

    def test_getyrdoy_1(self):
        '''Test the date to year, day of year code functionality'''
        date = pds.datetime(2009,1,1)
        yr, doy = pysat.utils.getyrdoy(date)
        assert ((yr == 2009) & (doy == 1))

    def test_getyrdoy_leap_year(self):
        '''Test the date to year, day of year code functionality (leap_year)'''
        date = pds.datetime(2008,12,31)
        yr, doy = pysat.utils.getyrdoy(date)
        assert ((yr == 2008) & (doy == 366)) 

    def test_custom_instrument_load(self):
        '''
        Test if the correct day is being loaded (End-to-End), 
        with no instrument file but routines are passed.
        '''
        import pysat.instruments.pysat_testing as test
        testInst = pysat.Instrument(inst_module=test, tag='', clean_level='clean')
        testInst.load(2009,32)
        assert testInst.date == pds.datetime(2009,2,1)	
        
    @raises(AttributeError)
    def test_custom_instrument_load_2(self):
        '''
        Test if an exception is thrown correctly if there is no 
        instrument file and supplied routines are incomplete.
        '''
        import pysat.instruments.pysat_testing as test
        del test.list_files
        testIn = pysat.Instrument(inst_module=test, tag='', clean_level='clean')
        testIn.load(2009,1)

    @raises(AttributeError)
    def test_custom_instrument_load_3(self):
        '''
        Test if an exception is thrown correctly if there is no 
        instrument file and supplied routines are incomplete.
        '''
        import pysat.instruments.pysat_testing as test
        del test.load
        testIn = pysat.Instrument(inst_module=test, tag='', clean_level='clean')
        testIn.load(2009,1)

    def test_data_padding(self):
        te = pysat.Instrument('pysat','testing', '', pad={'minutes':5})
        te.load(2009,1, verifyPad=True)
        assert ( (te.data.index[0] == te.date - pds.DateOffset(minutes=5)) & 
                (te.data.index[-1] == te.date + pds.DateOffset(hours=23,minutes=59,seconds=59) + 
                                        pds.DateOffset(minutes=5)) )
            
    def test_data_padding_removal(self):
        te = pysat.Instrument('pysat','testing', '', pad={'minutes':5})
        te.load(2009,1)
        assert (te.data.index[0] == te.date ) & \
                (te.data.index[-1] == te.date + pds.DateOffset(hour=23, minutes=59,seconds=59) )
        
    def test_basic_data_access_by_name(self):
        self.testInst.load(2009,1)
        assert np.all(self.testInst['uts'] == self.testInst.data['uts'])
        
    def test_data_access_by_row_slicing_and_name(self):
        self.testInst.load(2009,1)
        assert np.all(self.testInst[0:10,'uts'] == self.testInst.data.ix[0:10,'uts'])

    def test_data_access_by_row_and_name(self):
        self.testInst.load(2009,1)
        assert np.all(self.testInst[0,'uts'] == self.testInst.data.ix[0,'uts'])

    def test_data_access_by_datetime_and_name(self):
        self.testInst.load(2009,1)
        assert np.all(self.testInst[pysat.datetime(2009,1,1,0,0,0),'uts'] == self.testInst.data.ix[0,'uts'])
       
    def test_data_access_by_datetime_slicing_and_name(self):
        self.testInst.load(2009,1)
        assert np.all(self.testInst[pysat.datetime(2009,1,1,0,0,0):pysat.datetime(2009,1,1,0,0,10),'uts'] == 
                        self.testInst.data.ix[0:11,'uts'])
                        
    def test_setting_data_by_name(self):
        self.testInst.load(2009,1)
        self.testInst['doubleMLT'] = 2.*self.testInst['mlt']
        assert np.all(self.testInst['doubleMLT'] == 2.*self.testInst['mlt'])
        
    def test_set_bounds_by_date(self):
        start = pysat.datetime(2009,1,1)
        stop = pysat.datetime(2009,1,15)
        self.testInst.bounds = (start, stop)
        assert np.all(self.testInst._iter_list == pds.date_range(start, stop).tolist())

    def test_iterate_over_bounds_set_by_date(self):
        start = pysat.datetime(2009,1,1)
        stop = pysat.datetime(2009,1,15)
        self.testInst.bounds = (start, stop)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)            
        out = pds.date_range(start, stop).tolist()
        assert np.all(dates == out)
        
    def test_iterate_over_default_bounds(self):
        start = pysat.datetime(2008,1,1)
        stop = pysat.datetime(2010,12,31)
        self.testInst.bounds = (start, stop)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)            
        out = pds.date_range(start, stop).tolist()
        assert np.all(dates == out)
        
    def test_set_bounds_by_date_season(self):
        start = [pysat.datetime(2009,1,1), pysat.datetime(2009,2,1)]
        stop = [pysat.datetime(2009,1,15), pysat.datetime(2009,2,15)]
        self.testInst.bounds = (start, stop)
        out = pds.date_range(start[0], stop[0]).tolist()
        out.extend(pds.date_range(start[1], stop[1]).tolist())
        assert np.all(self.testInst._iter_list == out)

    def test_iterate_over_bounds_set_by_date_season(self):
        start = [pysat.datetime(2009,1,1), pysat.datetime(2009,2,1)]
        stop = [pysat.datetime(2009,1,15), pysat.datetime(2009,2,15)]
        self.testInst.bounds = (start, stop)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)            
        out = pds.date_range(start[0], stop[0]).tolist()
        out.extend(pds.date_range(start[1], stop[1]).tolist())
        assert np.all(dates == out)

    def test_set_bounds_by_fname(self):
        start = '01/01/09.nofile'
        stop = '01/03/09.nofile'
        self.testInst.bounds = (start, stop)
        assert np.all(self.testInst._iter_list == 
            ['01/01/09.nofile', '01/02/09.nofile', '01/03/09.nofile'])

    def test_iterate_over_bounds_set_by_fname(self):
        start = '01/01/09.nofile'
        stop = '01/15/09.nofile'
        start_d = pysat.datetime(2009,1,1)
        stop_d = pysat.datetime(2009,1,15)
        self.testInst.bounds = (start, stop)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)            
        out = pds.date_range(start_d, stop_d).tolist()
        assert np.all(dates == out)

    def test_set_bounds_by_fname_season(self):
        start = ['01/01/09.nofile', '02/01/09.nofile']
        stop = ['01/03/09.nofile', '02/03/09.nofile']
        self.testInst.bounds = (start, stop)
        assert np.all(self.testInst._iter_list == 
            ['01/01/09.nofile', '01/02/09.nofile', '01/03/09.nofile',
            '02/01/09.nofile', '02/02/09.nofile', '02/03/09.nofile'])

    def test_iterate_over_bounds_set_by_fname_season(self):
        start = ['01/01/09.nofile', '02/01/09.nofile']
        stop = ['01/15/09.nofile', '02/15/09.nofile']
        start_d = [pysat.datetime(2009,1,1), pysat.datetime(2009,2,1)]
        stop_d = [pysat.datetime(2009,1,15), pysat.datetime(2009,2,15)]
        self.testInst.bounds = (start, stop)
        dates = []
        for inst in self.testInst:
            dates.append(inst.date)            
        out = pds.date_range(start_d[0], stop_d[0]).tolist()
        out.extend(pds.date_range(start_d[1], stop_d[1]).tolist())
        assert np.all(dates == out)



import pysat
import pandas as pds
from nose.tools import assert_raises, raises
import nose.tools

from dateutil.relativedelta import relativedelta as relativedelta
class TestOrbits:
    
    def setup(self):
        '''Runs before every method to create a clean testing setup.'''
        info = {'index':'mlt'}
        self.testInst = pysat.Instrument('pysat','testing', '86400', 'clean',
                                        orbit_info=info)

    def teardown(self):
        '''Runs after every method to clean up previous testing.'''
        del self.testInst 
    
    def test_single_orbit_call_orbit_starts_0_UT_using_next(self):
        self.testInst.load(2009,1)
        self.testInst.orbits.next()
        ans = (self.testInst.data.index[0] == pds.datetime(2009,1,1))
        ans2 = (self.testInst.data.index[-1] == (pds.datetime(2009,1,1,1,36,59) ))
        # print (ans,ans2)
        # print (self.testInst.data.index[0], self.testInst.data.index[-1])
        assert ans & ans2

    def test_single_orbit_call_orbit_starts_0_UT_using_prev(self):
        self.testInst.load(2009,1)
        self.testInst.orbits.prev()
        ans = (self.testInst.data.index[0] == (pds.datetime(2009,1,1)-relativedelta(hours=1, minutes=37)) )
        ans2 = (self.testInst.data.index[-1] == (pds.datetime(2009,1,1)-relativedelta(seconds=1) ))
        assert ans & ans2

    def test_single_orbit_call_orbit_starts_off_0_UT_using_next(self):
        from dateutil.relativedelta import relativedelta as relativedelta
        self.testInst.load(2008,366)
        self.testInst.orbits.next()
        # print self.testInst.data.index[0], pds.datetime(2008,12,30, 23, 45), self.testInst.data.index[-1], (pds.datetime(2008,12,30, 23, 45)+relativedelta(hours=1, minutes=36, seconds=59) )
        ans = (self.testInst.data.index[0] == pds.datetime(2008,12,30, 23, 45))
        ans2 = (self.testInst.data.index[-1] == (pds.datetime(2008,12,30, 23, 45)+relativedelta(hours=1, minutes=36, seconds=59) ))
        assert ans & ans2

    def test_single_orbit_call_orbit_starts_off_0_UT_using_prev(self):
        self.testInst.load(2008,366)
        self.testInst.orbits.prev()
        ans = (self.testInst.data.index[0] == (pds.datetime(2008,12,30, 23, 45)-relativedelta(hours=1, minutes=37)) )
        ans2 = (self.testInst.data.index[-1] == (pds.datetime(2008,12,30, 23, 45)-relativedelta(seconds=1) ))
        assert ans & ans2

    def test_repeated_orbit_calls_symmetric_single_day_0_UT(self):
        self.testInst.load(2009,1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(10):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_multi_day_0_UT(self):
        self.testInst.load(2009,1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_single_day_off_0_UT(self):
        self.testInst.load(2008,366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(10):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)
	
    def test_repeated_orbit_calls_symmetric_multi_day_off_0_UT(self):
        self.testInst.load(2008,366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_antisymmetric_multi_day_off_0_UT(self):
        self.testInst.load(2008,366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        for j in range(10):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_antisymmetric_multi_multi_day_off_0_UT(self):
        self.testInst.load(2008,366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(40):
            self.testInst.orbits.prev()
        for j in range(20):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data )

    def test_repeated_orbit_calls_antisymmetric_multi_day_0_UT(self):
        self.testInst.load(2009,1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        for j in range(10):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data )

    def test_repeated_orbit_calls_antisymmetric_multi_multi_day_0_UT(self):
        self.testInst.load(2009,1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(40):
            self.testInst.orbits.prev()
        for j in range(20):
            self.testInst.orbits.next()
        print(control.data)
        print(self.testInst.data)
        assert all(control.data == self.testInst.data )


# -*- coding: utf-8 -*-
"""Supports the Vector Electric Field Instrument (VEFI) 
onboard the Communication and Navigation Outage Forecasting
System (C/NOFS) satellite. Downloads data from the
NASA Coordinated Data Analysis Web (CDAWeb).

Parameters
----------
platform : string
    'cnofs'
name : string
    'vefi'
tag : string
    Select measurement type, one of {'dc_b'}

Note
----
- tag = 'dc_b': 1 second DC magnetometer data

Warnings
--------
- Currently no cleaning routine.
- Module not written by VEFI team.
        
"""

from __future__ import print_function
from __future__ import absolute_import
import pandas as pds
import numpy as np
import pysat
import sys

import spacepy
from spacepy import pycdf
import pysat


def list_files(tag=None, data_path=None):
    """Return a Pandas Series of every file for chosen satellite data"""
    if data_path is not None:
        if tag == 'dc_b':
            return pysat.Files.from_os(data_path=data_path, 
            format_str='cnofs_vefi_bfield_1sec_{year:04d}{month:02d}{day:02d}_v05.cdf')
        else:
            raise ValueError('Unknown tag')
    else:
        raise ValueError ('A directory must be passed to the loading routine for VEFI')
            

def load(fnames, tag=None):
    if len(fnames) <= 0 :
        return pysat.DataFrame(None), None
    else:
        try:
            cdf = pycdf.CDF(fnames[0])
        except pycdf.CDFError:
            return pysat.DataFrame(), pysat.Meta()
            
        data = {}
        meta = pysat.Meta()
        for key in cdf.iterkeys():
            data[key] = cdf[key][...]
            try:             
                meta[key] = {'units':cdf[key].attrs['UNITS'],
                            'long_name':cdf[key].attrs['LABLAXIS'], 
                            'description':cdf[key].attrs['CATDESC']} 
                #meta[key] = {'description':cdf[key].attrs['VAR_NOTES']}          
            except KeyError:
                pass
        epoch = data.pop('Epoch')
        data = pysat.DataFrame(data, index=pds.to_datetime(epoch, unit='s'))
        return data, meta

def download(date_array, tag, data_path=None, user=None, password=None):
    """
    download vefi 1_second magnetic field data, layout consistent with pysat

    start and stop should be datetimes
    """
    import os
    import ftplib

    ftp = ftplib.FTP('cdaweb.gsfc.nasa.gov')   # connect to host, default port
    ftp.login()               # user anonymous, passwd anonymous@
    
    if tag == 'dc_b':
        ftp.cwd('/pub/data/cnofs/vefi/bfield_1sec')
    
        for date in date_array:
            fname = '{year1:4d}/cnofs_vefi_bfield_1sec_{year2:4d}{month:02d}{day:02d}_v05.cdf'
            fname = fname.format(year1=date.year, year2=date.year, month=date.month, day=date.day)
            local_fname = 'cnofs_vefi_bfield_1sec_{year:4d}{month:02d}{day:02d}_v05.cdf'.format(
                    year=date.year, month=date.month, day=date.day)
            saved_fname = os.path.join(data_path,local_fname) 
            try:
                print('Downloading file for '+date.strftime('%D'))
                sys.stdout.flush()
                ftp.retrbinary('RETR '+fname, open(saved_fname,'w').write)
            except ftplib.error_perm as exception:
                if exception[0][0:3] != '550':
                    raise
                else:
                    os.remove(saved_fname)
                    print('File not available for '+ date.strftime('%D'))



from __future__ import print_function
from __future__ import absolute_import

import os
import pandas as pds
# python 2/3 compatibility
try:
    basestring
except NameError:
    print ('setting basestring')
    basestring = str

#if 'basestring' not in globals():
    #print ('setting basestring')
    #basestring = str
    
from pysat import DataFrame, Series

class Meta(object):
    """
    Stores metadata for Instrument instance, similar to CF-1.6 netCDFdata standard.
    
    Parameters
    ----------
    metadata : pandas.DataFrame 
        DataFrame should be indexed by variable name that contains at minimum the 
        standard_name (name), units, and long_name for the data stored in the associated 
        pysat Instrument object.
        
    Attributes
    ----------
    data : pandas.DataFrame
        index is variable standard name, 'units' and 'long_name' are also stored along
        with additional user provided labels.
        
    """
    def __init__(self, metadata=None):
        self.replace(metadata=metadata)  

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False    

    def copy(self):
        from copy import deepcopy as deepcopy
        """Deep copy of the meta object."""
        return deepcopy(self) 
               
    def __setitem__(self, name, value):
        """Convenience method for adding metadata.
        
        Examples
        --------
        ::
        
            meta = pysat.Meta()
            meta['name'] = {'long_name':string, 'units':string}
            # update 'units' to new value
            meta['name'] = {'units':string}
            # update 'long_name' to new value
            meta['name'] = {'long_name':string}
            # attach new info with partial information, 'long_name' set to 'name2'
            meta['name2'] = {'units':string}
            # units are set to '' by default
            meta['name3'] = {'long_name':string}
        
        """
        
        if isinstance(value,dict):
            if 'units' not in value.keys():
                if name not in self.data.index:
                    value['units'] = ''
                else:
                    value['units'] = self.data.ix[name,'units']

            if 'long_name' not in value.keys():
                if name not in self.data.index:
                    # only set default if name not already in meta
                    value['long_name'] = name  
                else:
                    value['long_name'] = self.data.ix[name, 'long_name']
                
            # if hasattr(value[value.keys()[0]], '__iter__' ):
            if not isinstance(name, basestring): #hasattr(name, '__iter__' ):
                # an iterable of things         
                for key in value.keys():
                    if len(name) != len(value[key]):
                        raise ValueError('Length of names and inputs must be equal.')
                new = DataFrame(value, index=name)
                self.data = self.data.append(new) 
            else:
                for key in value.keys():
                    value[key] = [value[key]]
                new = DataFrame(value, index=[name])

                if name in self.data.index:
                    self.data = self.data.drop(name)
                    self.data = self.data.append(new)
                else:
                    self.data = self.data.append(new)
                
        if isinstance(value, Series):
            self.data.ix[name] = value
        
    def __getitem__(self,key):
        """Convenience method for obtaining metadata.
        
        Maps to pandas DataFrame.ix method.
        
        Examples
        --------
        ::
        
            print(meta['name'])
        
        """
        return self.data.ix[key]
        
    def replace(self, metadata=None):
        """Replace stored metadata with input data.
        
        Parameters
        ----------
        metadata : pandas.DataFrame 
            DataFrame should be indexed by variable name that contains at minimum the 
            standard_name (name), units, and long_name for the data stored in the associated 
            pysat Instrument object.
            
        """
        if metadata is not None:
            if isinstance(metadata, DataFrame):
                self.data = metadata
            else:
                raise ValueError("Input must be a pandas DataFrame type."+
                            "See other constructors for alternate inputs.")
            
        else:
            self.data = DataFrame(None, columns=['long_name', 'units'])
        #self._orig_data = self.data.copy()
        
    @classmethod
    def from_csv(cls, name=None, col_names=None, sep=None, **kwargs):
        """Create instrument metadata object from csv.
        
        Parameters
        ----------
        name : string
            absolute filename for csv file or name of file
            stored in pandas instruments location
        col_names : list-like collection of strings
            column names in csv and resultant meta object
        sep : string
            column seperator for supplied csv filename

        Note
        ----
        column names must include at least ['name', 'long_name', 'units'], 
        assumed if col_names is None.
        
        """
        import pysat
        req_names = ['name','long_name','units']
        if col_names is None:
            col_names = req_names
        elif not all([i in col_names for i in req_names]):
            raise ValueError('col_names must include name, long_name, units.')
            
        if sep is None:
            sep = ','
        
        if name is None:
            raise ValueError('Must supply an instrument name or file path.')
        elif not isinstance(name, str):
            raise ValueError('keyword name must be related to a string')               
        elif not os.path.isfile(name):
                # Not a real file, assume input is a pysat instrument name
                # and look in the standard pysat location.

                test =  os.path.join(pysat.__path__[0],'instruments',name)
                if os.path.isfile(test):
                    name = test
                else:
                    #trying to form an absolute path for success
                    test = os.path.abspath(name)
                    if not os.path.isfile(test):
                        raise ValueError("Unable to create valid file path.")
                    else:
                        #success
                        name = test
                                     
        mdata = pds.read_csv(name, names=col_names, sep=sep, **kwargs) 
        
        if not mdata.empty:
            # make sure the data name is the index
            mdata.index = mdata['name']
            del mdata['name']
            return cls(metadata=mdata)
        else:
            raise ValueError('Unable to retrieve information from ' + name)
        
    @classmethod
    def from_nc():
        """not implemented yet, load metadata from netCDF"""
        pass
        
    @classmethod
    def from_dict():
        """not implemented yet, load metadata from dict of items/list types"""
        pass



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import

import string
import os
import copy
import sys
import pandas as pds
import numpy as np

from . import _custom
from . import _files
from . import _orbits
from . import _meta
from . import utils
from pysat import data_dir
from pysat import DataFrame, Series


# main class for users
class Instrument(object):
    """Download, load, manage, modify and analyze science data.
    
    Parameters
    ----------    
    platform : string
        name of platform/satellite.
    name : string
        name of instrument. 
    tag : string, optional
        identifies particular subset of instrument data.
    inst_module : module, optional
        Provide instrument module directly. 
        Takes precedence over platform/name.
    clean_level : {'clean','dusty','dirty','none'}, optional
        level of data quality
    pad : pandas.DateOffset, or dictionary, optional
        Length of time to pad the begining and end of loaded data for 
        time-series processing. Extra data is removed after applying all 
        custom functions. Dictionary, if supplied, is simply passed to 
        pandas DateOffset.
    orbit_info : dict    
        Orbit information, {'index':index, 'kind':kind, 'period':period}.
        See pysat.Orbits for more information.            
    update_files : boolean, optional
        If True, query filesystem for instrument files and store. 
        files.get_new() will return no files after this call until 
        additional files are added.
    multi_file_day : boolean, optional (False by default)
        Set to True if Instrument data files for a day are spread across
        multiple files and data for day n could be found in a file
        with a timestamp of day n-1 or n+1.
        
    Attributes
    ----------
    data : pandas.DataFrame
        loaded science data 
    date : pandas.datetime
        date for loaded data
    yr : int
        year for loaded data
    bounds : (datetime/filename/None, datetime/filename/None)
        bounds for loading data, supply array_like for a season with gaps
    doy : int
        day of year for loaded data
    files : pysat.Files
        interface to instrument files
    meta : pysat.Meta
        interface to instrument metadata, similar to netCDF 1.6
    orbits : pysat.Orbits
        interface to extracting data orbit-by-orbit
    custom : pysat.Custom
        interface to instrument nano-kernel
    kwargs : dictionary
        keyword arguments passed to instrument loading routine
    
    Note
    ----
    Pysat attempts to load the module platform_name.py located in
    the pysat/instruments directory. This module provides the underlying 
    functionality to download, load, and clean instrument data. 
    Alternatively, the module may be supplied directly
    using keyword inst_module.
    
    Examples
    --------     
    :: 
           
        # 1-second mag field data
        vefi = pysat.Instrument(platform='cnofs', 
                                name='vefi', 
                                tag='dc_b', 
                                clean_level='clean')
        start = pysat.datetime(2009,1,1)
        stop = pysat.datetime(2009,1,2)
        vefi.download(start, stop)
        vefi.load(date=start)
        print(vefi['dB_mer'])
        print(vefi.meta['db_mer'])
    
        # 1-second thermal plasma parameters
        ivm = pysat.Instrument(platform='cnofs', 
                                name='ivm', 
                                tag='', 
                                clean_level='clean')
        ivm.download(start,stop)
        ivm.load(2009,1)
        print(ivm['ionVelmeridional'])
        
        # Ionosphere profiles from GPS occultation
        cosmic = pysat.Instrument('cosmic2013', 
                                    'gps', 
                                    'ionprf', 
                                    altitude_bin=3)
        # bins profile using 3 km step
        cosmic.download(start, stop, user=user, password=password)
        cosmic.load(date=start)

    """

    def __init__(self, platform=None, name=None, tag=None, clean_level='clean',
                 update_files=False, pad=None, orbit_info=None,
                 inst_module=None, multi_file_day=False, *arg, **kwargs):

        if inst_module is None:
            if isinstance(platform, str) and isinstance(name, str):              
                self.platform = platform.lower() 
                self.name = name.lower() 
                self._assign_funcs(by_name=True)
            elif (platform is None) and (name is None):
                # creating "empty" Instrument object with this path
                self.name = ''
                self.platform = ''
                self._assign_funcs()                
            else:
                raise ValueError('Inputs platform and name must both be strings, or both None.')                        
        else:
            # user expected to have provided module
            try:
                self.name = inst_module.name.lower()
                self.platform = inst_module.platform.lower()
            except AttributeError:
                raise AttributeError(string.join(('A name and platform attribute for the ',
                                     'instrument is required if supplying routine module directly.')))
            self._assign_funcs(inst_module=inst_module)
            
        self.tag = tag.lower() if tag is not None else ''    
        self.clean_level = clean_level.lower() if clean_level is not None else 'none'

        self.data = DataFrame(None)
        self.meta = _meta.Meta()
        # function processing class, processes data on load
        self.custom = _custom.Custom()
        # create arrays to store data around loaded day
        # enables padding across day breaks with minimal loads
        self._next_data = DataFrame(None)
        self._next_data_track = []
        self._prev_data = DataFrame(None)
        self._prev_data_track = []
        self._curr_data = DataFrame(None)
        
        # arguments for padding
        if isinstance(pad, pds.DateOffset):
            self.pad = pad
        elif isinstance(pad, dict):
            self.pad = pds.DateOffset(**pad)
        elif pad is None:
            self.pad = None
        else:
            raise ValueError('pad must be a dictionary or a pandas.DateOffset instance.')

        # multi file day
        self.multi_file_day = multi_file_day

        self.kwargs = kwargs

        # load file list function, which returns dict of files
        # as well as data start and end dates
        self.files = _files.Files(self)
        if update_files:
            self.files.refresh(store=True)
        # set bounds for iteration based upon data properties
        # setting (None,None) loads default bounds
        self.bounds = (None, None)
        self.date = None
        self._fid = None
        self.yr = None
        self.doy = None
        self._load_by_date = False

        # initialize orbit support
        if orbit_info is None:
            orbit_info = {'index': None, 'kind': None, 'period': None}
        self.orbits = _orbits.Orbits(self, **orbit_info)
        # run instrument init function, a basic pass function is used
        # if user doesn't supply the init function
        self._init_rtn(self)

    def __getitem__(self, key): 
        """
        Convenience notation for accessing data; inst['name'] is inst.data.name
        
        Examples
        --------
        ::
        
            # By name
            inst['name']
            # By position  
            inst[row_index, 'name']  
            # Slicing by row 
            inst[row1:row2, 'name']            
            # By Date  
            inst[datetime, 'name']
            # Slicing by date, inclusive
            inst[datetime1:datetime2, 'name']
            # Slicing by name and row/date  
            inst[datetime1:datetime1, 'name1':'name2']
            
        """

        if isinstance(key, tuple):
            # support slicing
            return self.data.ix[key[0], key[1]]
        else:
            return self.data[key]

    def __setitem__(self, key, new):
        """Convenience method for adding data to instrument.
        
        Examples
        --------
        ::
        
            # Simple Assignment, default metadata assigned
            # 'long_name' = 'name'
            # 'units' = ''
            inst['name'] = newData
            # Assignment with Metadata 
            inst['name'] = {'data':new_data, 
                            'long_name':long_name, 
                            'units':units}
        
        Note
        ----
        If no metadata provided and if metadata for 'name' not already stored 
        then default meta information is also added, 
        long_name = 'name', and units = ''.
        
        """
        if isinstance(new, dict):
            # metadata should be included in dict
            self.data[key] = new.pop('data')
            # pass the rest to meta
            self.meta[key] = new
        else:
            if isinstance(key, tuple):
                self.data.ix[key[0], key[1]] = new
                if key[1] not in self.meta.data.index:
                    self.meta[key[1]] = {'long_name': key, 'units': ''}
            elif isinstance(key, str):
                self.data[key] = new  
                if key not in self.meta.data.index:   
                    # add in default metadata because none was supplied
                    self.meta[key] = {'long_name': key, 'units': ''}
            elif isinstance(new, DataFrame):
                self.data[key] = new[key]
                for ke in key:
                    if ke not in self.meta.data.index:   
                        # add in default metadata because none was supplied
                        self.meta[ke] = {'long_name': ke, 'units': ''}
            else:
                raise ValueError("No support for supplied input key")           
    
    def copy(self):
        """Deep copy of the entire Instrument object."""
        return copy.deepcopy(self)
    
    def _pass_func(*args, **kwargs):
        pass     

    def _assign_funcs(self, by_name=False, inst_module=None):        
        """Assign all external science instrument methods to Instrument object."""
        import importlib
        # set defaults
        self._list_rtn = self._pass_func
        self._load_rtn = self._pass_func
        self._default_rtn = self._pass_func
        self._clean_rtn = self._pass_func
        self._init_rtn = self._pass_func
        self._download_rtn = self._pass_func
            
        if by_name: 
            # look for code with filename name, any errors passed up
            inst = importlib.import_module(''.join(('.', self.platform, '_',
                                           self.name)), package='pysat.instruments')
        elif inst_module is not None:
            # user supplied an object with relevant instrument routines
            inst = inst_module
        else:
            # no module or name info, default pass functions assigned
            return

        try:
            self._load_rtn = inst.load
            self._list_rtn = inst.list_files
            self._download_rtn = inst.download
        except AttributeError:
            raise AttributeError(string.join(('A load, file_list, and download routine',
                                 'are required for every instrument.')))
        try:
            self._default_rtn = inst.default
        except AttributeError:
            pass
        try:
            self._init_rtn = inst.init
        except AttributeError:
            pass   
        try:
            self._clean_rtn = inst.clean
        except AttributeError:
            pass

        return

    def _load_data(self, date=None, fid=None):
        """
        Load data for an instrument on given date or fid, dependng upon input.
        
        """

        if fid is not None:
            # get filename based off of index value
            # remove direct access of data
            fname = self.files[fid]
        elif date is not None:
            fname = self.files[date: date+pds.DateOffset(days=1)]
        else:
            raise ValueError('Must supply either a date or file id number.')
   
        if len(fname) > 0:    
            load_fname = [os.path.join(self.files.data_path, f) for f in fname]
            data, mdata = self._load_rtn(load_fname, tag=self.tag, **self.kwargs)
        else:
            data = DataFrame(None)
            mdata = _meta.Meta()

        if not data.empty: 
            if not isinstance(data, DataFrame):
                raise TypeError(string.join(('Data returned by instrument load',
                                'routine must be a pandas.DataFrame')))
            if not isinstance(mdata, _meta.Meta):
                raise TypeError('Metadata returned must be a pysat.Meta object')
            if date is not None:
                print(' '.join(('Returning', self.platform, self.name, self.tag, 'data for',
                                   date.strftime('%D'))))
            else:
                if len(fname) == 1:
                    # this check was zero
                    print(' '.join(('Returning', self.platform, self.name, self.tag,
                                       'data from', fname[0])))
                else:
                    print(' '.join(('Returning', self.platform, self.name, self.tag,
                                       'data from', fname[0], '::', fname[-1])))
        else:
            # no data signal
            print(' '.join(('No', self.platform, self.name, self.tag, 'data for',
                               date.strftime('%D'))))
    
        return data, mdata
        
    def _load_next(self):
        """Load the next days data (or file) without incrementing the date.
        Repeated calls will not advance date/file and will produce the same data
        
        Uses info stored in object to either increment the date, 
        or the file. Looks for self._load_by_date flag. 
         
        """
        if self._load_by_date:
            next_date = self.date + pds.DateOffset(days=1)
            return self._load_data(date=next_date)
        else:
            return self._load_data(fid=self._fid+1)

    def _load_prev(self):
        """Load the next days data (or file) without decrementing the date.
        Repeated calls will not decrement date/file and will produce the same data
        
        Uses info stored in object to either decrement the date, 
        or the file. Looks for self._load_by_date flag.  
        
        """

        if self._load_by_date:
            prev_date = self.date - pds.DateOffset(days=1)
            return self._load_data(date=prev_date)
        else:
            return self._load_data(fid=self._fid-1)

    def load(self, yr=None, doy=None, date=None, fname=None, fid=None, 
             verifyPad=False):
        """Load instrument data into Instrument object .data.

        Parameters
        ----------
        yr : integer
            year for desired data
        doy : integer
            day of year
        date : datetime object
            date to load
        fname : 'string'
            filename to be loaded
        verifyPad : boolean 
            if True, padding data not removed (debug purposes)

        Note
        ----
        Loads data for a chosen instrument into .data. Any functions chosen
        by the user and added to the custom processing queue (.custom.add)
        are automatically applied to the data before it is available to 
        user in .data.
        
        """

        if date is not None:
            # date supplied getyrdoy checks if it is datetime
            year, doy = utils.getyrdoy(date)
            self.yr = year 
            self.doy = doy 
            self.date = date
            self._fid = None
            self._load_by_date = True
            inc = pds.DateOffset(days=1)
            curr = date
        elif (yr is not None) & (doy is not None):
            # if date not defined but both yr and doy are
            self.date = pds.datetime(yr,1,1) + pds.DateOffset(days=(doy-1))
            self.yr = yr
            self.doy = doy
            self._fid = None
            self._load_by_date = True
            inc = pds.DateOffset(days=1)
            curr = self.date
        elif fname is not None:
            # date will have to be set later by looking at the data
            self.date = None
            self.yr = None
            self.doy = None
            self._load_by_date = False
            # if no index, called func tries to find file in instrument dir,
            # throws IOError if it fails
            self._fid = self.files.get_index(fname)
            inc = 1
            curr = self._fid.copy()
        elif fid is not None:
            self._load_by_date = False	    
            self._fid = fid
            self.date = None
            self.yr = None
            self.doy = None
            inc = 1
            curr = fid
        else:
            raise TypeError('Must supply a yr,doy pair, or datetime object, ' +
                            'or filename to load data from.')

        self.orbits._reset()
        # if pad is true, need to have a three day/file load
        if (self.pad is not None) | self.multi_file_day:
            if self._next_data.empty & self._prev_data.empty:
                # data has not already been loaded for previous and next days
                # load data for all three
                print('Initializing three day/file window')
                # using current date or fid
                self._prev_data, self._prev_meta = self._load_prev()
                self._curr_data, self._curr_meta = self._load_data(date=self.date, fid=self._fid)
                self._next_data, self._next_meta = self._load_next()
            else:
                # moving forward in time
                if self._next_data_track == curr:
                    self._prev_data, self._prev_meta = self._curr_data, self._curr_meta
                    self._curr_data, self._curr_meta = self._next_data, self._next_meta
                    self._next_data, self._next_meta = self._load_next()
                # moving backward in time
                elif self._prev_data_track == curr:
                    self._next_data, self._next_meta = self._curr_data, self._curr_meta
                    self._curr_data, self._curr_meta = self._prev_data, self._prev_meta
                    self._prev_data, self._prev_meta = self._load_prev()
                # jumped in time/or switched from filebased to date based access
                else:
                    self._prev_data, self._prev_meta = self._load_prev()
                    self._curr_data, self._curr_meta = self._load_data(date=self.date, fid=self._fid)   
                    self._next_data, self._next_meta = self._load_next()

            # make tracking indexes consistent with new loads
            self._next_data_track = curr + inc
            self._prev_data_track = curr - inc
            # attach data to object
            if not self._curr_data.empty :
                self.data = self._curr_data.copy()
                self.meta = self._curr_meta.copy()
            else:
                self.data = DataFrame(None)
                # line below removed as it would delete previous meta, if any
                # if you end a seasonal analysis with a day with no data, then no meta
                # self.meta = _meta.Meta()

            if self.multi_file_day:
                self.data = self.data[self.date : self.date+pds.DateOffset(hours=23, minutes=59, seconds=59, nanoseconds=99999999)]
            # pad data based upon passed parameter

            if (not self._prev_data.empty) & (not self.data.empty):
                if self.multi_file_day and self._load_by_date:
                    padLeft = self._prev_data[(self.date):self._curr_data.index[0]]
                else:
                    padLeft = self._prev_data[(self._curr_data.index[0]-self.pad):self._curr_data.index[0]]
                self.data = pds.concat([padLeft[0:-1], self.data])

            if (not self._next_data.empty) & (not self.data.empty):

                if self.multi_file_day and self._load_by_date:
                    padRight = self._next_data[self.date : (self.date+pds.DateOffset(hours=23, minutes=59, seconds=59, nanoseconds=99999999))]
                else:
                    padRight = self._next_data[self._curr_data.index[-1]:(self._curr_data.index[-1]+self.pad)]
                self.data = pds.concat([self.data, padRight[1:]])
                
        # if self.pad is False, load single day
        else:
            self.data, meta = self._load_data(date=self.date, fid=self._fid) 
            if not self.data.empty:
                self.meta = meta   
               
        # check if load routine actually returns meta
        if self.meta.data.empty:
            self.meta[self.data.columns] = {'long_name': self.data.columns,
                                            'units': ['']*len(self.data.columns)}
        if not self.data.empty:
            self._default_rtn(self)
        # clean
        if (not self.data.empty) & (self.clean_level != 'none'):
            self._clean_rtn(self)   
        # apply custom functions
        if not self.data.empty:
            self.custom._apply_all(self)
        # remove the excess padding, if any applied
        if (self.pad is not None)& (not self.data.empty) & (not verifyPad) :
            self.data = self.data[self._curr_data.index[0]:self._curr_data.index[-1]]
        # if loading by file set the yr, doy, and date
        if not self._load_by_date:
            temp = self.data.index[0]
            temp = pds.datetime(temp.year, temp.month, temp.day)
            self.date = temp
            self.yr, self.doy = utils.getyrdoy(self.date)

        sys.stdout.flush()
        return

    def download(self, start, stop, freq='D', user=None, password=None):
        """Download data for given Instrument object from start to stop.
        
        Parameters
        ----------
        start : pandas.datetime
            start date to download data
        stop : pandas.datetime
            stop date to download data
        freq : string
            Stepsize between dates for season, 'D' for daily, 'M' monthly (see pandas)
        user : string
            username, if required by instrument data archive
        password : string
            password, if required by instrument data archive
            
        Note
        ----
        Data will be downloaded to pysat_data_dir/patform/name/tag
        
        If Instrument bounds are set to defaults they are updated
        after files are downloaded.
        
        """
        import errno
        # make sure directories are there, otherwise create them
        try:
            os.mkdir(os.path.join(data_dir, self.platform))
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise    
        try:
            os.mkdir(os.path.join(data_dir, self.platform, self.name))
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise    
        try:
            os.mkdir(self.files.data_path)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise
        date_array = utils.season_date_range(start, stop, freq=freq)
        if user is None:
            self._download_rtn(date_array,
                               tag=self.tag,
                               data_path=self.files.data_path)
        else:
            self._download_rtn(date_array,
                               tag=self.tag,
                               data_path=self.files.data_path,
                               user=user,
                               password=password)
        # get current file date range
        first_date = self.files.start_date
        last_date = self.files.stop_date
            
        print('Updating pysat file list')
        self.files.refresh(store=True)

        # if instrument object has default bounds, update them
        if len(self.bounds[0]) == 1:
            if (self.bounds[0][0] == first_date) and (self.bounds[1][0] == last_date):
                print('Updating instrument object bounds.')
                self.bounds = None
        
    @property
    def bounds(self):
        """Boundaries for iterating over instrument object by date or file.

        Parameters
        ----------
        start : datetime object, filename, or None (default)
            start of iteration, if None uses first data date.
            list-like collection also accepted
        end :  datetime object, filename, or None (default)
                end of iteration, inclusive. If None uses last data date.
                list-like collection also accepted

        Note
        ----
        Both start and stop must be the same type (date, or filename) or None

        Examples
        --------
        ::

            inst = pysat.Instrument(platform=platform,
                                    name=name,
                                    tag=tag)
            start = pysat.datetime(2009,1,1)
            stop = pysat.datetime(2009,1,31)
            inst.bounds = (start,stop)

            start2 = pysat.datetetime(2010,1,1)
            stop2 = pysat.datetime(2010,2,14)
            inst.bounds = ([start, start2], [stop, stop2])

        """
        return self._iter_start, self._iter_stop
    
    @bounds.setter        
    def bounds(self, value=None):
        if value is None:
            value = (None, None)
        if len(value) < 2:
            raise ValueError('Must supply both a start and end date/file' +
                             'Supply None if you want the first/last possible')

        start = value[0]
        end = value[1]
        # get the frequency, or step size, of season
        if len(value) == 3:
            step = value[2]
        else:
            # default do daily
            step = 'D'

        if (start is None) and (end is None):
            # set default
            self._iter_start = [self.files.start_date]
            self._iter_stop = [self.files.stop_date]
            self._iter_type = 'date'
            if self._iter_start[0] is not None:
                # check here in case Instrument is initialized with no input
                self._iter_list = utils.season_date_range(self._iter_start, self._iter_stop, freq=step)
        # elif (not isinstance(start, str)) and (not isinstance(end, str)):
        elif (hasattr(start, '__iter__') and not isinstance(start,str)) and (hasattr(end, '__iter__') and not isinstance(end,str)):
            base = type(start[0])
            for s, t in zip(start, end):
                if (type(s) != type(t)) or (type(s) != base):
                    raise ValueError('Start and end items must all be of the same type')
            if isinstance(start[0], str):
                self._iter_type = 'file'
                self._iter_list = self.files.get_file_array(start, end)
            elif isinstance(start[0], pds.datetime):
                self._iter_type = 'date'
                self._iter_list = utils.season_date_range(start, end, freq=step)
            else:
                raise ValueError('Input is not a known type, string or datetime')
            self._iter_start = start
            self._iter_stop = end
        # elif (not isinstance(start, str)) or (not isinstance(end, str)):
        elif (hasattr(start, '__iter__') and not isinstance(start,str)) or (hasattr(end, '__iter__') and not isinstance(end,str)):
            raise ValueError('Both start and end must be iterable if one bound is iterable')

        elif isinstance(start, str) or isinstance(end, str):
            if isinstance(start, pds.datetime) or isinstance(end, pds.datetime):
                raise ValueError('Not allowed to mix file and date bounds')
            if start is None:
                start = self.files[0]
            if end is None:
                end = self.files.files[-1]
            self._iter_start = [start]
            self._iter_stop = [end]
            self._iter_list = self.files.get_file_array(self._iter_start, self._iter_stop)
            self._iter_type = 'file'

        elif isinstance(start, pds.datetime) or isinstance(end, pds.datetime):
            if start is None:
                start = self.files.start_date
            if end is None:
                end = self.files.stop_date
            self._iter_start = [start]
            self._iter_stop = [end]
            self._iter_list = utils.season_date_range(start, end, freq=step)
            self._iter_type = 'date'
        else:
            raise ValueError('Provided an invalid combination of bounds. ' +
                             'if specifying by file, both bounds must be by file. Other ' +
                             'combinations of datetime objects and None are allowed.')

    def __iter__(self):
        """Iterates instrument object by loading subsequent days or files.

        Note
        ----
        Limits of iteration, and iteration type (date/file)
        set by `bounds` attribute.

        Default bounds are the first and last dates from files on local system.

        Examples
        --------
        ::

            inst = pysat.Instrument(platform=platform,
                                    name=name,
                                    tag=tag)
            start = pysat.datetime(2009,1,1)
            stop = pysat.datetime(2009,1,31)
            inst.bounds = (start,stop)
            for inst in inst:
            print('Another day loaded', inst.date)

        """

        if self._iter_type == 'file':
            for fname in self._iter_list:
                self.load(fname=fname)
                yield self       
 
        elif self._iter_type == 'date':
            for date in self._iter_list:
                self.load(date=date)
                yield self            
                
    def next(self):
        """Manually iterate through the data loaded in Instrument object.
        
        Bounds of iteration and iteration type (day/file) are set by 
        `bounds` attribute.
        
        Note
        ----
        If there were no previous calls to load then the 
        first day(default)/file will be loaded.
         
        """
        
        if self._iter_type == 'date':
            if self.date is not None:
                idx, = np.where(self._iter_list == self.date)
                if (len(idx) == 0) | (idx+1 >= len(self._iter_list)):
                    raise StopIteration('Outside the set date boundaries.')
                else:
                    idx += 1
                    self.load(date=self._iter_list[idx[0]])
            else:
                self.load(date=self._iter_list[0])

        elif self._iter_type == 'file':
            if self._fid is not None:
                idx, = np.where(self._iter_list == self._fid)
                if (len(idx) == 0) | (idx+1 >= len(self._iter_list)):
                    raise StopIteration('Outside the set file boundaries.')
                else:
                    idx += 1
                    self.load(fid=self._iter_list[idx[0]])
            else:
                self.load(fid=self._iter_list[0])

    def prev(self):
        """Manually iterate backwards through the data in Instrument object.
        
        Bounds of iteration and iteration type (day/file) 
        are set by `bounds` attribute.
        
        Note
        ----
        If there were no previous calls to load then the 
        first day(default)/file will be loaded.
            
        """
        
        if self._iter_type == 'date':
            if self.date is not None:
                idx, = np.where(self._iter_list == self.date)
                if (len(idx) == 0) | (idx-1 < 0):
                    raise StopIteration('Outside the set date boundaries.')
                else:
                    idx -= 1
                    self.load(date=self._iter_list[idx[0]])
            else:
                self.load(date=self._iter_list[-1])

        elif self._iter_type == 'file':
            if self._fid is not None:
                idx, = np.where(self._iter_list == self._fid)
                if (len(idx) == 0) | (idx-1 < 0):
                    raise StopIteration('Outside the set file boundaries.')
                else:
                    idx -= 1
                    self.load(fid=self._iter_list[idx[0]])
            else:
                # last file by default
                self.load(fid=self._iter_list[-1])

    def to_netcdf3(self, fname=None):
        """Stores loaded data into a netCDF3 64-bit file.
        
        Parameters
        ----------
        fname : string
            full path to save instrument object to
        
        Note
        ----
        Stores 1-D data along dimension 'time' - the date time index.
        
        Stores object data (e.g. dataframes within series) separately
                    
         - The name of the series is used to prepend extra variable
           dimensions within netCDF, key_2, key_3; first dimension time
         - The index organizing the data stored as key_sample_index
         - from_netcdf3 uses this naming scheme to reconstruct data structure
            
        The datetime index is stored as 'UNIX time'. netCDF-3 doesn't support
        64-bit integers so it is stored as a 64-bit float. This results in a
        loss of datetime precision when converted back to datetime index
        up to hundreds of nanoseconds. Use netCDF4 if this is a problem.
          
        All attributes attached to instrument meta are written to netCDF attrs.
        
        """
        
        import netCDF4
        
        with netCDF4.Dataset(fname, mode='w', format='NETCDF3_64BIT') as out_data:
        
            num = len(self.data.index)
            out_data.createDimension('time', num)
            
            # write out the datetime index
            cdfkey = out_data.createVariable('unix_time', 'f8', dimensions=('time'),)
            cdfkey.units = 's'
            cdfkey.long_name = 'UNIX time'
            cdfkey.calendar = 'noleap'
            cdfkey[:] = self.data.index.astype(int)*1.E-9
            # store all of the data names in meta            
            for key in self.meta.data.index:
                if self[key].dtype != np.dtype('O'):
                    # not an object, simple column of data, write it out
                    cdfkey = out_data.createVariable(key, 
                                                     self[key].dtype,
                                                     dimensions=('time'), )
                    cdfkey.units = self.meta[key].units
                    cdfkey.long_name = self.meta[key].long_name
                    cdfkey[:] = self[key].values
                else:
                    # we are dealing with a more complicated object
                    # presuming a series with a dataframe in each location
                    dims = np.shape(self[key].iloc[0])
                    obj_dim_names = []
                    # don't need to recreate last dimension, 
                    # it covers number of columns
                    for i, dim in enumerate(dims[:-1]):
                        obj_dim_names.append(key+'_dim_%i' % (i+1))
                        out_data.createDimension(obj_dim_names[-1], dim)
                    var_dim = tuple(['time']+obj_dim_names)
                    
                    # iterate over columns and store
                    for col in self[key].iloc[0].columns:
                        coltype = self[key].iloc[0][col].dtype
                        if coltype == np.int64:
                            coltype = np.int32
                        cdfkey = out_data.createVariable(col, 
                                                         coltype,
                                                         dimensions=var_dim)
                        cdfkey.long_name = col
                        cdfkey.units = ''
                        for i in xrange(num):
                            cdfkey[i, :] = self[key].iloc[i][col].values.astype(coltype)
                            
                    # store the dataframe index for each time of main dataframe
                    datetime_flag = False
                    coltype = self[key].iloc[0].index.dtype
                    # check for datetime index
                    if coltype == np.dtype('<M8[ns]'):
                        coltype = 'f8'
                        datetime_flag = True
                    
                    cdfkey = out_data.createVariable(key+'_sample_index', 
                                                     coltype, dimensions=var_dim)
                    if datetime_flag:
                        cdfkey.units = 's'
                        cdfkey.long_name = 'UNIX time'
                        for i in xrange(num):
                            cdfkey[i, :] = self[key].iloc[i].index.astype(int)*1.E-9
                    else:
                        cdfkey.units = ''
                        cdfkey.long_name = key
                        for i in xrange(num):
                            cdfkey[i, :] = self[key].iloc[i].index.to_native_types()

                    if self[key].iloc[0].index.name is not None:
                        cdfkey.long_name = self[key].iloc[0].index.name
                    
            # store any non standard attributes
            base_attrb = dir(Instrument())
            this_attrb = dir(self)
            
            adict = {}
            for key in this_attrb:
                if key not in base_attrb:
                    if key[0] != '_':
                        adict[key] = self.__getattribute__(key)
            # store any non-standard attributes attached to meta
            base_attrb = dir(_meta.Meta())
            this_attrb = dir(self.meta)
            for key in this_attrb:
                if key not in base_attrb:
                    if key[0] != '_':
                        adict[key] = self.__getattribute__(key)
            adict['pysat_version'] = 1.0
            adict['Conventions'] = 'CF-1.6'
            out_data.setncatts(adict)



from __future__ import print_function
from __future__ import absolute_import

import numpy as np
import pandas as pds
from pysat import Series, DataFrame


class Orbits(object):
    """Determines orbits on the fly and provides orbital data in .data.
    
    Determines the locations of orbit breaks in the loaded data in inst.data
    and provides iteration tools and convenient orbit selection via
    inst.orbit[orbit num].
    
    Parameters
    ----------
    sat : pysat.Instrument instance
        instrument object to determine orbits for
    index : string
        name of the data series to use for determing orbit breaks
    kind : {'local time', 'longitude', 'polar'}
        kind of orbit, determines how orbital breaks are determined
        
        - local time: negative gradients in lt or breaks in inst.data.index
        - longitude: negative gradients or breaks in inst.data.index
        - polar: zero crossings in latitude or breaks in inst.data.index
    period : np.timedelta64
        length of time for orbital period, used to gauge when a break
        in the datetime index (inst.data.index) is large enough to
        consider it a new orbit
            
    Note
    ----
    class should not be called directly by the user, use the interface provided
    by inst.orbits where inst = pysat.Instrument()
    
    Warning
    -------
    This class is still under development.
    
    Examples
    --------
    ::
        
        info = {'index':'longitude', 'kind':'longitude'}
        vefi = pysat.Instrument(platform='cnofs', name='vefi', tag='dc_b', 
                                clean_level=None, orbit_info=info)
        start = pysat.datetime(2009,1,1)
        stop = pysat.datetime(2009,1,10)
        vefi.load(date=start)
        vefi.bounds(start, stop)
        
        # iterate over orbits
        for vefi in vefi.orbits:
            print('Next available orbit ', vefi['dB_mer'])
                            
        # load fifth orbit of first day
        vefi.load(date=start)
        vefi.orbits[5]
        
        # less convenient load
        vefi.orbits.load(5)
        
        # manually iterate orbit
        vefi.orbits.next()
        # backwards
        vefi.orbits.prev()
           
    """

    def __init__(self, sat=None, index=None, kind=None, period=None):
        # create null arrays for storing orbit info
        if sat is None:
            raise ValueError('Must provide a pysat instrument object when initializing ' +
                             'orbits class.')
        else:
            # self.sat = weakref.proxy(sat)
            self.sat = sat

        if kind is None:
            kind = 'local time'
        else:
            kind = kind.lower()

        if period is None:
            period = np.timedelta64(97, 'm')
        self.orbit_period = period

        if (kind == 'local time') or (kind == 'lt'):
            self._detBreaks = self._equaBreaks
        elif (kind == 'longitude') or (kind == 'long'):
            self._detBreaks = self._longBreaks
        elif kind == 'polar':
            self._detBreaks = self._polarBreaks
        else:
            raise ValueError('Unknown kind of orbit requested.')

        self._orbit_breaks = []
        self.num = []
        self.current = 0
        self.orbit_index = index

    def __getitem__(self, key):
        """Enable convenience notation for loading orbit into parent object.
        
        Examples
        --------
        ::
        
            inst.load(date=date)
            inst.orbits[4]
            print('Orbit data ', inst.data)
        
        Note
        ----
        A day of data must already be loaded.
            
        """

        self.load(key)

    def _reset(self):
        # create null arrays for storing orbit info
        self._orbit_breaks = []
        self.num = None
        self.current = 0

    def _calcOrbits(self):
        """Prepares data structure for breaking data into orbits. Not intended for end user."""
        # if the breaks between orbit have not been defined, define them
        # also, store the data so that grabbing different orbits does not
        # require reloads of whole dataset
        if self._orbit_breaks == []:
            # determine orbit breaks
            self._detBreaks()
            # store a copy of data
            self._fullDayData = self.sat.data.copy()
            # set current orbit counter to zero (default)
            self.current = 0

    def _equaBreaks(self):
        """Determine where breaks in an equatorial satellite orbit occur.
        
        Looks for negative gradients in local time (or longitude) as well as 
        breaks in UT.

        """

        if self.orbit_index is None:
            try:
                self.sat['slt']
                self.orbit_index = 'slt'
            except ValueError:
                try:
                    self.sat['mlt']
                    self.orbit_index = 'mlt'
                except ValueError:
                    try:
                        self.sat['glong']
                        self.orbit_index = 'glong'
                    except ValueError:
                        raise ValueError('Unable to find a valid index (slt/mlt/glong) for determining orbits.')
        else:
            try:
                self.sat[self.orbit_index]
            except ValueError:
                raise ValueError('Provided orbit index does not exist in loaded data')

        lt_diff = self.sat[self.orbit_index].diff()
        ut_vals = Series(self.sat.data.index)
        ut_diff = ut_vals.diff()
        # locations where derivative is less than 0
        # or, look for breaks because the length of time between samples is too large
        # deriv of datetime index is in naneseconds
        ind, = np.where((lt_diff < -0.1))
        if len(ind) > 0:
            ind = np.hstack((ind, np.array([len(self.sat[self.orbit_index])])))
            # look at distance between breaks
            dist = ind[1:] - ind[0:-1]
            # only keep orbit breaks with a distance greater than 1
            # done for robustness
            if len(ind) > 1:
                if min(dist) == 1:
                    print('There are orbit breaks right next to each other')
                ind = ind[dist > 1]

            # check for large positive gradients around the break that would
            # suggest not a true orbit break, but rather bad orbit_index values
            new_ind = []
            for idx in ind:
                tidx, = np.where(lt_diff[idx - 5:idx + 6] > 0.1)

                if len(tidx) != 0:
                    for tidx in tidx:
                        # look at time change vs local time change
                        if ut_diff[idx - 5:idx + 6].iloc[tidx] < lt_diff[idx - 5:idx + 6].iloc[
                            tidx] / 24. * self.orbit_period:
                            # change in ut inconsistent with change in local time
                            # increases in local time require a change in ut
                            pass
                        else:
                            new_ind.append(idx)
                            break
                else:
                    new_ind.append(idx)

            ind = np.array(new_ind)

        # now, look for breaks because the length of time between samples is too large, thus there is no break in slt/mlt/etc
        ut_ind, = np.where((ut_diff > self.orbit_period) | ((ut_diff / self.orbit_period > (np.abs(lt_diff / 24.))) & (
        ut_diff / self.orbit_period > 0.95)))  # & lt_diff.notnull() ))# & (lt_diff != 0)  ) )   #added the or and check after or on 10/20/2014
        if len(ut_ind) > 0:
            ind = np.hstack((ind, ut_ind))
            ind = np.sort(ind)
            ind = np.unique(ind)
            print('Time Gap')

        # now that most problems in orbits should have been caught, look at
        # the time difference between orbits
        orbit_ut_diff = ut_vals[ind].diff()
        orbit_lt_diff = self.sat[self.orbit_index][ind].diff()

        idx, = np.where((orbit_ut_diff / self.orbit_period - orbit_lt_diff.values / 24.) > 0.97)
        idx = np.hstack((0, idx))
        if len(ind) > 0:
            ind = ind[idx]

        # number of orbits
        num_orbits = len(ind) + 1
        # create orbitbreak index
        ind = np.hstack((np.array([0]), ind))
        # set index of orbit breaks
        self._orbit_breaks = ind
        # set number of orbits for the day
        self.num = num_orbits

    def _longBreaks(self):
        """
        Determine where breaks in a satellite orbit occur, based upon
        changes in longitude.
        
        Looks for negative gradients in longitude as well as 
        breaks in UT.

        """
        if self.orbit_index is None:
            try:
                self.sat['glong']
                self.orbit_index = 'glong'
            except ValueError:
                raise ValueError(''.join(('Unable to find a valid index',
                                          'for determining orbits. Provide one in orbit_index')))
        else:
            try:
                self.sat[self.orbit_index]
            except ValueError:
                raise ValueError('Provided orbit index does not exist in loaded data')

        lt_diff = self.sat[self.orbit_index].diff()
        ut_vals = Series(self.sat.data.index)
        ut_diff = ut_vals.diff()
        # locations where derivative is less than 0
        # or, look for breaks because the length of time between samples is too large
        # deriv of datetime index is in naneseconds
        ind, = np.where((lt_diff < -0.1))
        if len(ind) > 0:
            ind = np.hstack((ind, np.array([len(self.sat[self.orbit_index])])))
            # look at distance between breaks
            dist = ind[1:] - ind[0:-1]
            # only keep orbit breaks with a distance greater than 1
            # done for robustness, though this could rarely introduce a problem
            if len(ind) > 1:
                if min(dist) == 1:
                    print('There are orbit breaks right next to each other')
                ind = ind[dist > 1]

            # check for large positive gradients around the break that would
            # suggest not a true orbit break, but rather bad orbit_index values
            new_ind = []
            for idx in ind:
                tidx, = np.where(lt_diff[idx - 5:idx + 6] > 0.1)

                if len(tidx) != 0:
                    for tidx in tidx:
                        # look at time change vs longitude change
                        if ut_diff[idx - 5:idx + 6].iloc[tidx] < lt_diff[idx - 5:idx + 6].iloc[
                            tidx] / 360. * self.orbit_period:
                            # change in ut inconsistent with change in local time
                            # increases in longitude require a change in ut
                            pass
                        # print 'Fake Orbit Break.'
                        else:
                            new_ind.append(idx)
                            break
                else:
                    new_ind.append(idx)

            ind = np.array(new_ind)

        # now, look for breaks because the length of time between samples is so large,
        # that there is no break in slt/mlt/etc
        ut_ind, = np.where(
            (ut_diff > self.orbit_period) |
            ((ut_diff / self.orbit_period > (np.abs(lt_diff / 360.))) &
             (ut_diff / self.orbit_period > 0.95))
        )
        #  & lt_diff.notnull() ))# & (lt_diff != 0)  ) )   #added the or and check after or on 10/20/2014

        if len(ut_ind) > 0:
            ind = np.hstack((ind, ut_ind))
            ind = np.sort(ind)
            ind = np.unique(ind)
            print('Time Gap')

        # now that most problems in orbits should have been caught, look at
        # the time difference between orbits
        orbit_ut_diff = ut_vals[ind].diff()
        orbit_lt_diff = self.sat[self.orbit_index][ind].diff()

        idx, = np.where((orbit_ut_diff / self.orbit_period - orbit_lt_diff.values / 360.) > 0.97)
        idx = np.hstack((0, idx))
        if len(ind) > 0:
            ind = ind[idx]

        # number of orbits
        num_orbits = len(ind) + 1
        # create orbitbreak index
        ind = np.hstack((np.array([0]), ind))
        # set index of orbit breaks
        self._orbit_breaks = ind
        # set number of orbits for the day
        self.num = num_orbits

    def _polarBreaks(self):
        """Determine where breaks in a polar orbiting satellite orbit occur.
        
        Looks for sign changes in latitude (magnetic or geographic) as well as 
        breaks in UT.

        """

        if self.orbit_index is None:
            raise ValueError('Orbit properties must be defined at ' +
                             'pysat.Instrument object instantiation. See Instrument docs.')
        else:
            try:
                self.sat[self.orbit_index]
            except ValueError:
                raise ValueError('Provided orbit index does not appear to exist in loaded data')

        # determine where orbit index goes from positive to negative
        pos = self.sat[self.orbit_index] >= 0
        npos = -pos
        change = (pos.values[:-1] & npos.values[1:]) | (npos.values[:-1] & pos.values[1:])

        ind, = np.where(change)
        ind += 1
        if len(ind) > 0:
            ind = np.hstack((ind, np.array([len(self.sat[self.orbit_index])])))
            # look at distance between breaks
            dist = ind[1:] - ind[0:-1]
            # only keep orbit breaks with a distance greater than 1
            # done for robustness
            if len(ind) > 1:
                if min(dist) == 1:
                    print('There are orbit breaks right next to each other')
                ind = ind[dist > 1]

        ut_diff = Series(self.sat.data.index).diff()
        ut_ind, = np.where(ut_diff / self.orbit_period > 0.95)

        if len(ut_ind) > 0:
            ind = np.hstack((ind, ut_ind))
            ind = np.sort(ind)
            ind = np.unique(ind)
            # print 'Time Gap'
        # number of orbits
        num_orbits = len(ind) + 1
        # create orbitbreak index
        ind = np.hstack((np.array([0]), ind))
        # set index of orbit breaks
        self._orbit_breaks = ind
        # set number of orbits for the day
        self.num = num_orbits

    def _getBasicOrbit(self, orbit=None):
        """Load a particular orbit into .data for loaded day.

        Parameters
        ----------
        orbit : int
            orbit number, 1 indexed, negative indexes allowed, -1 last orbit

        Note
        ----
        A day of data must be loaded before this routine functions properly.
        If the last orbit of the day is requested, it will NOT automatically be
        padded with data from the next day.
        
        """
        # ensure data exists
        if len(self.sat.data) > 0:
            # ensure proper orbit metadata present
            self._calcOrbits()

            # ensure user is requesting a particular orbit
            if orbit is not None:
                # pull out requested orbit
                if orbit == -1:
                    # load orbit data into data
                    self.sat.data = self._fullDayData[self._orbit_breaks[self.num + orbit]:]
                    self.current = self.num + orbit + 1
                elif ((orbit < 0) & (orbit >= -self.num)):
                    # load orbit data into data
                    self.sat.data = self._fullDayData[
                                    self._orbit_breaks[self.num + orbit]:self._orbit_breaks[self.num + orbit + 1]]
                    self.current = self.num + orbit + 1
                elif (orbit < self.num) & (orbit != 0):
                    # load orbit data into data
                    self.sat.data = self._fullDayData[self._orbit_breaks[orbit - 1]:self._orbit_breaks[orbit]]
                    self.current = orbit
                elif orbit == self.num:
                    self.sat.data = self._fullDayData[self._orbit_breaks[orbit - 1]:]
                    self.current = orbit  # recent addition, wondering why it wasn't there before, could just be a bug
                elif orbit == 0:
                    raise ValueError('Orbits start at 1, 0 not allowed')
                else:
                    # gone too far
                    self.sat.data = []
                    raise ValueError('Requested an orbit past total orbits for day')
            else:
                raise ValueError('Must set an orbit')

    def load(self, orbit=None):
        """Load a particular orbit into .data for loaded day.

        Parameters
        ----------
        orbit : int
            orbit number, 1 indexed

        Note
        ----    
        A day of data must be loaded before this routine functions properly.
        If the last orbit of the day is requested, it will automatically be
        padded with data from the next day. The orbit counter will be 
        reset to 1.
        
        """
        if len(self.sat.data) > 0:  # ensure data exists
            # set up orbit metadata
            self._calcOrbits()
            # ensure user supplied an orbit
            if orbit is not None:
                # pull out requested orbit
                if orbit < 0:
                    # negative indexing consistent with numpy, -1 last, -2 second
                    # to last, etc.
                    orbit = self.num + 1 + orbit

                if orbit == 1:
                    # change from orig copied from _core, didn't look correct.
                    # self._getBasicOrbit(orbit=2)
                    try:
                        true_date = self.sat.date  # .copy()

                        self.sat.prev()
                        # if and else added becuase of CINDI turn off 6/5/2013,turn on 10/22/2014
                        # crashed when starting on 10/22/2014
                        # prev returned empty data
                        if len(self.sat.data) > 0:
                            self.load(orbit=-1)
                        else:
                            self.sat.next()
                            self._getBasicOrbit(orbit=1)
                        # check that this orbit should end on the current day
                        delta = pds.to_timedelta(true_date - self.sat.data.index[0])
                        # print 'checking if first orbit should land on requested day'
                        # print self.sat.date, self.sat.data.index[0], delta, delta >= self.orbit_period
                        # print delta - self.orbit_period
                        if delta >= self.orbit_period:
                            # the orbit loaded isn't close enough to date
                            # to be the first orbit of the day, move forward
                            self.next()
                    except StopIteration:
                        # print 'going for basic orbit'
                        self._getBasicOrbit(orbit=1)
                        print('Loaded Orbit:%i' % self.current)
                        # check if the first orbit is also the last orbit

                elif orbit == self.num:
                    # we get here if user asks for last orbit
                    # make sure that orbit data goes across daybreak as needed
                    # load previous orbit
                    if self.num != 1:
                        self._getBasicOrbit(self.num - 1)
                        self.next()
                    else:
                        self._getBasicOrbit(orbit=-1)

                elif orbit < self.num:
                    # load orbit data into data
                    self._getBasicOrbit(orbit)
                    print('Loaded Orbit:%i' % self.current)

                else:
                    # gone too far
                    self.sat.data = DataFrame()
                    raise Exception('Requested an orbit past total orbits for day')
            else:
                raise Exception('Must set an orbit')
        else:
            print('No data loaded in instrument object to determine orbits.')

    def next(self, *arg, **kwarg):
        """Load the next orbit into .data.

        Note
        ----
        Forms complete orbits across day boundaries. If no data loaded
        then the first orbit from the first date of data is returned.
            
        """

        # first, check if data exists
        if len(self.sat.data) > 0:
            # set up orbit metadata
            self._calcOrbits()

            # if current orbit near the last, must be careful
            if self.current == (self.num - 1):
                # first, load last orbit data
                self._getBasicOrbit(orbit=-1)
                # End of orbit may occur on the next day
                load_next = True
                if self.sat._iter_type == 'date':
                    delta = pds.to_timedelta(self.sat.date - self.sat.data.index[-1]) + np.timedelta64(1, 'D')
                    if delta >= self.orbit_period:
                        # don't need to load the next day because this orbit ends more than a orbital
                        # period from the next date
                        load_next = False

                if load_next:
                    # the end of the user's desired orbit occurs tomorrow, need to form a complete orbit
                    # save this current orbit, load the next day, combine data, select the correct orbit
                    temp_orbit_data = self.sat.data.copy()
                    try:
                        # loading next day/file clears orbit breaks info
                        self.sat.next()
                        if len(self.sat.data) > 0:
                            # combine this next day's data with previous last orbit, grab the first one
                            self.sat.data = pds.concat(
                                [temp_orbit_data[:self.sat.data.index[0] - pds.DateOffset(microseconds=1)],
                                 self.sat.data])
                            self._getBasicOrbit(orbit=1)
                        else:
                            # no data, go back a day and grab the last orbit. As complete as orbit can be
                            self.sat.prev()
                            self._getBasicOrbit(orbit=-1)
                    except StopIteration:
                        pass
                    del temp_orbit_data
                print('Loaded Orbit:%i' % self.current)

            elif self.current == (self.num):
                # at the last orbit, need to be careful about getting the next orbit
                # save this current orbit and load the next day
                temp_orbit_data = self.sat.data.copy()
                # load next day, which clears orbit breaks info
                self.sat.next()
                # combine this next day orbit with previous last orbit to ensure things are correct
                if len(self.sat.data) > 0:
                    pad_next = True
                    # check if data padding is really needed, only works when loading by date
                    if self.sat._iter_type == 'date':
                        delta = pds.to_timedelta(self.sat.date - temp_orbit_data.index[-1])
                        if delta >= self.orbit_period:
                            # the end of the previous orbit is more than an orbit away from today
                            # we don't have to worry about it
                            pad_next = False
                    if pad_next:
                        # orbit went across day break, stick old orbit onto new data and grab second orbit (first is old)
                        self.sat.data = pds.concat(
                            [temp_orbit_data[:self.sat.data.index[0] - pds.DateOffset(microseconds=1)], self.sat.data])
                        # select second orbit of combined data
                        self._getBasicOrbit(orbit=2)
                    else:
                        # padding from the previous orbit wasn't needed, can just grab the first orbit of loaded data
                        self._getBasicOrbit(orbit=1)
                        if self.sat._iter_type == 'date':
                            delta = pds.to_timedelta(self.sat.date + pds.DateOffset(days=1) - self.sat.data.index[0])
                            if delta < self.orbit_period:
                                # this orbits end occurs on the next day
                                # though we grabbed the first orbit, missing data means the first available
                                # orbit in the data is actually the last for the day
                                # Resetting to second to last orbit and then calling next()
                                # will get the last orbit, accounting for tomorrow's data as well.
                                self.current = self.num - 1
                                self.next()
                else:
                    # no data for the next day
                    # continue loading data until there is some
                    # nextData raises StopIteration when it reaches the end, leaving this function
                    while len(self.sat.data) == 0:
                        self.sat.next()
                    self._getBasicOrbit(orbit=1)

                del temp_orbit_data
                print('Loaded Orbit:%i' % self.current)

            elif self.current == 0:
                # no current orbit set, grab the first one
                # using load command to specify the first orbit
                # which automatically loads prev day if needed to form complete orbit
                self.load(orbit=1)

            elif self.current < (self.num - 1):
                # since we aren't close to the last orbit, just pull the next orbit
                self._getBasicOrbit(orbit=self.current + 1)
                print('Loaded Orbit:%i' % self.current)

            else:
                raise Exception(
                    'You ended up where nobody should ever be. Talk to someone about this fundamental failure.')

        else:  # no data
            while len(self.sat.data) == 0:
                # keep going until data is found
                # next raises stopIteration at end of data set, no more data possible
                self.sat.next()
            # we've found data, grab the next orbit
            self.next()

    def prev(self, *arg, **kwarg):
        """Load the next orbit into .data.

        Note
        ----
        Forms complete orbits across day boundaries. If no data loaded
        then the last orbit of data from the last day is loaded into .data.
            
        """

        # first, check if data exists
        if len(self.sat.data) > 0:
            # set up orbit metadata
            self._calcOrbits()
            # if not close to the first orbit,just pull the previous orbit

            if (self.current > 2) & (self.current <= self.num):
                # load orbit and put it into self.sat.data
                self._getBasicOrbit(orbit=self.current - 1)

            # if current orbit near the last, must be careful
            elif self.current == 2:
                # first, load prev orbit data
                self._getBasicOrbit(orbit=self.current - 1)
                # need to save this current orbit and load the next day
                temp_orbit_data = self.sat.data[self.sat.date:]
                # load previous day, which clears orbit breaks info

                try:
                    self.sat.prev()
                    # combine this next day orbit with previous last orbit
                    if len(self.sat.data) > 0:
                        self.sat.data = pds.concat([self.sat.data, temp_orbit_data])
                        # select first orbit of combined data
                        self._getBasicOrbit(orbit=-1)
                    else:
                        self.sat.next()
                        self._getBasicOrbit(orbit=1)
                except StopIteration:
                    # if loading the first orbit, of first day of data, you'll
                    # end up here as the attempt to make a full orbit will
                    # move the date backwards, and StopIteration is made.
                    # everything is already ok, just move along
                    pass

                del temp_orbit_data

            elif self.current < 2:
                # first, load next orbit data
                self._getBasicOrbit(orbit=1)
                # need to save this current orbit and load the next day
                temp_orbit_data = self.sat.data[self.sat.date:]
                # load previous day, which clears orbit breaks info
                self.sat.prev()
                # combine this next day orbit with previous last orbit
                if len(self.sat.data) > 0:
                    self.sat.data = pds.concat([self.sat.data, temp_orbit_data])
                    # select first orbit of combined data
                    self._getBasicOrbit(orbit=-2)

                del temp_orbit_data

            else:
                raise Exception(
                    'You ended up where noone should ever be. Talk to someone about this fundamental failure.')
            print('Loaded Orbit:%i' % self.current)
        else:
            # no data
            while len(self.sat.data) == 0:
                self.sat.prev()  # raises stopIteration at end of dataset
            self.prev()

    def __iter__(self):
        """Support iteration by orbit.
        
        For each iteration the next available orbit is loaded into
        inst.data.
        
        Examples
        --------
        ::
        
            for inst in inst.orbits:
                print 'next available orbit ', inst.data
        
        Note
        ----
        Limits of iteration set by setting inst.bounds.

        """
        while True:
            self.next()
            yield self.sat



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
"""
Instrument independent seasonal averaging routine. Supports averaging
1D and 2D data.
"""

import pysat
import numpy as np
import pandas as pds
import collections

def median2D(inst, bin1, label1, bin2, label2, data_label, 
             returnData=False):
    """Return a 2D average of data_label over a season and label1, label2.

    Parameters
    ----------
        bin#: [min, max, number of bins]
        label#: string 
            identifies data product for bin#
        data_label: list-like 
            contains strings identifying data product(s) to be averaged

    Returns
    -------
    median : dictionary
        2D median accessed by data_label as a function of label1 and label2
        over the season delineated by bounds of passed instrument objects.
        Also includes 'count' and 'avg_abs_dev' as well as the values of
        the bin edges in 'bin_x' and 'bin_y'.
    
    """

    # create bins
    binx = np.linspace(bin1[0], bin1[1], bin1[2]+1)
    biny = np.linspace(bin2[0], bin2[1], bin2[2]+1)

    numx = len(binx)-1
    numy = len(biny)-1
    numz = len(data_label)

    # create array to store all values before taking median
    yarr = np.arange(numy)
    xarr = np.arange(numx)
    zarr = np.arange(numz)
    ans = [ [ [collections.deque() for i in xarr] for j in yarr] for k in zarr]

    # set up output arrays
    medianAns = [ [ [ None for i in xarr] for j in yarr] for k in zarr]
    countAns = [ [ [ None for i in xarr] for j in yarr] for k in zarr]
    devAns = [ [ [ None for i in xarr] for j in yarr] for k in zarr]    

    # do loop to iterate over instrument season
    for inst in inst:
        # collect data in bins for averaging
        if len(inst.data) != 0:
            xind = np.digitize(inst.data[label1], binx)-1
            for xi in xarr:
                xindex, = np.where(xind==xi)
                if len(xindex) > 0:
                    yData = inst.data.iloc[xindex]
                    yind = np.digitize(yData[label2], biny)-1
                    for yj in yarr:
                        yindex, = np.where(yind==yj)
                        if len(yindex) > 0:
                            for zk in zarr:
                                ans[zk][yj][xi].extend( yData.ix[yindex,data_label[zk]].tolist() )

    # all of the loading and storing data is done
    # determine what kind of data is stored
    # if just numbers, then use numpy arrays to store data
    # if the data is a more generalized object, use lists to store data
    # need to find first bin with data
    dataType = [None for i in np.arange(numz)]
    for zk in zarr:
        breakNow=False
        for yj in yarr:
            for xi in xarr:
                if len(ans[zk][yj][xi]) > 0:
                    dataType[zk] = type(ans[zk][yj][xi][0]) 
                    breakNow = True
                    break 
            if breakNow:
                break

    # determine if normal number objects are being used or if there
    # are more complicated objects
    objArray = np.zeros(len(zarr))
    objArray = [False]*len(zarr)
    for i,thing in enumerate(dataType):
         if thing == pds.core.series.Series:
            objArray[i] = 'Series'
         if thing == pds.core.frame.DataFrame:
            objArray[i] = 'Frame'

    objArray = np.array(objArray)

    # if some pandas data series are returned in average, return a list
    objidx, = np.where(objArray == 'Series')
    if len(objidx) > 0:
        for zk in zarr[objidx]:
            for yj in yarr:
                for xi in xarr:
                    if len(ans[zk][yj][xi]) > 0:
                        ans[zk][yj][xi] = list(ans[zk][yj][xi])
                        medianAns[zk][yj][xi] =  pds.DataFrame(ans[zk][yj][xi] ).median(axis=0)
                        countAns[zk][yj][xi] = len(ans[zk][yj][xi])
                        devAns[zk][yj][xi] = pds.DataFrame([abs(temp) - medianAns[zk][yj][xi] for temp in ans[zk][yj][xi] ] ).median(axis=0)
                                                                    
    # if some pandas DataFrames are returned in average, return a list
    objidx, = np.where(objArray == 'Frame')
    if len(objidx) > 0:
        for zk in zarr[objidx]:
            for yj in yarr:
                for xi in xarr:                    
                    if len(ans[zk][yj][xi]) > 0:
                        ans[zk][yj][xi] = list(ans[zk][yj][xi])
                        countAns[zk][yj][xi] = len(ans[zk][yj][xi])
                        test = pds.Panel.from_dict(dict([(i,temp) for i,temp in enumerate(ans[zk][yj][xi]) ]) )
                        medianAns[zk][yj][xi] = test.median(axis=0)
                        devAns[zk][yj][xi] = (test.subtract(medianAns[zk][yj][xi], axis=0)).abs().median(axis=0, skipna=True)
                                                                                                      
    objidx, = np.where((objArray == False) | (objArray == 'False'))
    if len(objidx) > 0:
        for zk in zarr[objidx]:
            medianAns[zk] = np.zeros((numy, numx))*np.nan
            countAns[zk] = np.zeros((numy, numx))*np.nan
            devAns[zk] = np.zeros((numy, numx))*np.nan
            for yj in yarr:
                for xi in xarr:
                    # convert deque storing data into numpy array
                    ans[zk][yj][xi] = np.array(ans[zk][yj][xi])
                    # filter out an NaNs in the arrays
                    idx, = np.where(np.isfinite(ans[zk][yj][xi]))
                    ans[zk][yj][xi] = (ans[zk][yj][xi])[idx]
                    # perform median averaging
                    if len(idx) > 0:
                        medianAns[zk][yj,xi] = np.median(ans[zk][yj][xi])
                        countAns[zk][yj,xi] = len(ans[zk][yj][xi])
                        devAns[zk][yj,xi] = np.median(abs(ans[zk][yj][xi] - medianAns[zk][yj,xi]))

    # prepare output
    output = {}
    for i,label in enumerate(data_label):
        output[label] = {'median': medianAns[i], 
                        'count':countAns[i],
                        'avg_abs_dev':devAns[i],
                        'bin_x': binx,
                        'bin_y': biny}

        if returnData:
            output[label]['data'] = ans[i]

    return output

# simple averaging through multiple iterations

def mean_by_day(inst, data_label):
    """Mean of data_label by day over Instrument.bounds

    Parameters
    ----------
    data_label : string
        string identifying data product to be averaged

    Returns
    -------
    mean : pandas Series
        simple mean of data_label indexed by day

    """
    return _core_mean(inst, data_label, by_day=True)

def mean_by_orbit(inst, data_label):
    """Mean of data_label by orbit over Instrument.bounds

    Parameters
    ----------
    data_label : string
        string identifying data product to be averaged

    Returns
    -------
    mean : pandas Series
        simple mean of data_label indexed by start of each orbit

    """
    return _core_mean(inst, data_label, by_orbit=True)

def mean_by_file(inst, data_label):
    """Mean of data_label by orbit over Instrument.bounds

    Parameters
    ----------
    data_label : string
        string identifying data product to be averaged

    Returns
    -------
    mean : pandas Series
        simple mean of data_label indexed by start of each file

    """
    return _core_mean(inst, data_label, by_file=True)

def _core_mean(inst, data_label, by_orbit=False, by_day=False, by_file=False):

    if by_orbit:
        iterator = inst.orbits
    elif by_day or by_file:
        iterator = inst
    else:
        raise ValueError('A choice must be made, by day, file, or orbit')

    # create empty series to hold result
    mean_val = pds.Series()
    # iterate over season, calculate the mean
    for inst in iterator:
        if not inst.data.empty:
                # compute mean absolute using pandas functions and store
                # data could be an image, or lower dimension, account for 2D and lower
                data = inst[data_label]
                data.dropna(inplace=True)

                if by_orbit or by_file:
                   date = inst.data.index[0]
                else:
                   date = inst.date

                if isinstance(data.iloc[0], pds.DataFrame):
                    data_panel = pds.Panel.from_dict(dict([(i,data.iloc[i]) for i in xrange(len(data))]))
                    mean_val[date] = data_panel.mean(axis=0,skipna=True)
                elif isinstance(data.iloc[0], pds.Series):
                    data_frame = pds.DataFrame(data.tolist())
                    data_frame.index = data.index
                    mean_val[date] = data_frame.mean(axis=0, skipna=True)
                else:
                    mean_val[date] = inst[data_label].mean(axis=0,skipna=True)
    del iterator
    return mean_val
